#  Copyright 2021, Yahoo
#  Licensed under the terms of the Apache 2.0 license. See the LICENSE file in the project root for terms
import yaml


# https://github.com/yaml/pyyaml/issues/234#issuecomment-765894586
class Dumper(yaml.Dumper):
    def increase_indent(self, flow=False, *args, **kwargs):
        return super().increase_indent(flow=flow, indentless=False)



#  Copyright 2021, Yahoo
#  Licensed under the terms of the Apache 2.0 license. See the LICENSE file in the project root for terms
from abc import ABC
from argparse import Namespace
from pathlib import Path
from typing import Any, Optional

from ..core.executor.MachineTargetExecutor import MachineTargetExecutor
from ..core.executor.SelfTargetExecutor import SelfTargetExecutor
from ..testplan.attack import TargetType
from . import YChaosCLIHook, YChaosTestplanInputSubCommand


class YChaosCLITargetExecutorHook(YChaosCLIHook, ABC):
    def __init__(self, app):
        super(YChaosCLITargetExecutorHook, self).__init__(app=app)
        self._exitcode = 0


class Execute(YChaosTestplanInputSubCommand):

    name = "execute"
    help = "The execute subcommand of YChaos"

    def __init__(self, **kwargs):
        super(Execute, self).__init__(**kwargs)

        self.test_plan_path: Path = kwargs.pop("testplan")

        self.testplan = self.get_validated_test_plan(self.test_plan_path)
        if not self.testplan:
            self.set_exitcode(1)
            return

        self.executor: Optional[Any] = None

    # Section: Hooks for each Target Type

    def _register_target_hooks(self) -> None:
        """
        Registers all the CLI hooks for target executor
        Returns:
            None
        """
        assert isinstance(self.executor, (MachineTargetExecutor, SelfTargetExecutor))
        target_type = self.executor.__target_type__

        class OnTargetExecutorStart(YChaosCLITargetExecutorHook):
            def __call__(self):
                self.console.log(f"Starting attack. executor={target_type}")

        class OnTargetUnreachableHook(YChaosCLITargetExecutorHook):
            def __call__(self, result):
                self.console.print(f"{result._host.get_name()} is unreachable")
                self._exitcode = 1

        class OnTargetFailedHook(YChaosCLITargetExecutorHook):  # pragma: no cover
            def __call__(self, result):
                self.console.log(
                    f"Attack on {result._host.get_name()} failed. Task={result.task_name}"
                )
                self._exitcode = 1

        class OnNoTargetsFoundHook(YChaosCLITargetExecutorHook):
            def __call__(self):
                self.console.print("No targets found for attack. Bailing out..")

        self.executor.register_hook("on_start", OnTargetExecutorStart(self.app))
        self.executor.register_hook(
            "on_target_unreachable", OnTargetUnreachableHook(self.app)
        )
        self.executor.register_hook("on_target_failed", OnTargetFailedHook(self.app))

        if self.testplan.attack.target_type != TargetType.SELF:
            self.executor.register_hook(
                "on_no_targets_found", OnNoTargetsFoundHook(self.app)
            )

    def build_executor(self):
        if self.testplan.attack.target_type == TargetType.MACHINE:
            self.executor = MachineTargetExecutor(testplan=self.testplan)

        elif self.testplan.attack.target_type == TargetType.SELF:
            self.executor = SelfTargetExecutor(testplan=self.testplan)

        else:
            raise NotImplementedError()

        self._register_target_hooks()

    def run(self):
        self.executor.execute()

        for hook_name, hooks in self.executor.hooks.items():
            if any([_h._exitcode != 0 for _h in hooks]):
                self.set_exitcode(1)
                break

    @classmethod
    def main(cls, args: Namespace) -> Any:
        execute = cls(**vars(args))

        if execute._exitcode != 0:
            return execute._exitcode

        execute.build_executor()
        execute.run()

        return execute._exitcode



#  Copyright 2021, Yahoo
#  Licensed under the terms of the Apache 2.0 license. See the LICENSE file in the project root for terms

from .subparsers import SubCommand, SubCommandParsersAction

__all__ = ["SubCommandParsersAction", "SubCommand"]



#  Copyright 2021, Yahoo
#  Licensed under the terms of the Apache 2.0 license. See the LICENSE file in the project root for terms
from pathlib import Path
from typing import Optional, Union

from pydantic import BaseModel


class ApplicationSettings(BaseModel):
    """
    Defines the Global Settings that are consistent in both Development &
    Production scenarios
    """

    APP = "YChaos"
    APP_DESC = "YChaos, The resilience testing framework"
    PROG = "ychaos"

    COMMAND_IDENTIFIER = "_cmd.{}"
    LOG_FILE_PATH: Optional[Path] = None

    @classmethod
    def get_instance(cls):
        return cls()

    @classmethod
    def get_version(cls):
        import pkg_resources

        return pkg_resources.get_distribution(cls.get_instance().PROG).version


class DevSettings(ApplicationSettings):
    """
    Defines the Development settings for YChaos Application.
    """

    CONFIG = "dev"


class ProdSettings(DevSettings):
    """
    Defines the Production settings for YChaos Application

    Prod Settings overrides the Dev Settings class and redefines all
    the constants defined in DevSettings that can be used in the production scenario
    """

    CONFIG = "prod"


class Settings:
    __instance: Optional[Union[DevSettings, ProdSettings]] = None

    @classmethod
    def get_instance(cls) -> Union[DevSettings, ProdSettings]:
        if cls.__instance is None:
            cls(config="prod")
        assert cls.__instance is not None
        return cls.__instance

    def __init__(self, config):
        if config == "dev":
            self.__class__.__instance = DevSettings()
        elif config == "prod":
            self.__class__.__instance = ProdSettings()
        else:
            raise AttributeError("Unknown configuration found")



#  Copyright 2021, Yahoo
#  Licensed under the terms of the Apache 2.0 license. See the LICENSE file in the project root for terms

from enum import Enum
from typing import Any, Dict

from pydantic import BaseModel


class SystemState(Enum):
    STEADY = "STEADY"
    CHAOS = "CHAOS"
    RECOVERED = "RECOVERED"


class SchemaModel(BaseModel):
    class Config:
        @staticmethod
        def schema_extra(schema: Dict[str, Any], model) -> None:
            for prop in schema.get("properties", {}).values():
                prop.pop("title", None)



#  Copyright 2021, Yahoo
#  Licensed under the terms of the Apache 2.0 license. See the LICENSE file in the project root for terms

from typing import Any, Dict


class YChaosError(Exception):
    def __init__(self, e, message, **kwargs):
        self.exception = e
        self.message: str = message
        self.attrs: Dict[str, Any] = kwargs



#  Copyright 2021, Yahoo
#  Licensed under the terms of the Apache 2.0 license. See the LICENSE file in the project root for terms
import importlib.util
from pathlib import Path
from typing import Any, Dict

from pydantic import Field, PrivateAttr

from .agent import Agent, AgentConfig


class ContribAgentConfig(AgentConfig):
    name = "contrib"
    path: Path = Field(..., description="The path of the agent python file")

    agent_class: str = Field(
        default="AgentClass", description="The class name of the contributed Agent"
    )
    agent_config_class: str = Field(
        default="AgentConfigClass",
        description="The class name of the contributed Agent config",
    )

    config: Dict[Any, Any] = Field(
        default=dict(),
        description="The configuration that will be passed to the community agent",
    )

    # For storing imported module
    # Initialized late in the _import_module() method
    _module: Any = PrivateAttr()

    def __init__(self, **kwargs):
        super(ContribAgentConfig, self).__init__(**kwargs)
        self._import_module()

        # Validate that `config` adheres to the schema of the contrib agent
        self.config = self.get_agent_config_class()(**self.config)

    def _import_module(self):
        specification = importlib.util.spec_from_file_location(
            name="", location=self.path
        )
        self._module = importlib.util.module_from_spec(specification)

        assert self._module is not None
        assert specification.loader is not None
        specification.loader.exec_module(module=self._module)  # type: ignore

    def get_agent_class(self) -> Any:
        agent_klass = getattr(self._module, self.agent_class)
        assert issubclass(agent_klass, Agent)
        return agent_klass

    def get_agent_config_class(self) -> Any:
        agent_config_klass = getattr(self._module, self.agent_config_class)
        assert issubclass(agent_config_klass, AgentConfig)
        return agent_config_klass

    def get_agent(self):
        return self.get_agent_class()(self.config)



#  Copyright 2021, Yahoo
#  Licensed under the terms of the Apache 2.0 license. See the LICENSE file in the project root for terms
import shlex
import subprocess  # nosec: using shlex

from pydantic import validate_arguments

from ....testplan.verification import PythonModuleVerification
from ..data import VerificationData, VerificationStateData
from .BaseVerificationPlugin import BaseVerificationPlugin


class PythonModuleVerificationPlugin(BaseVerificationPlugin):

    __verification_type__ = "python_module"

    @validate_arguments
    def __init__(self, config: PythonModuleVerification, state_data: VerificationData):
        super(PythonModuleVerificationPlugin, self).__init__(config, state_data)

    def run_verification(self) -> VerificationStateData:
        _command = [self.config.executable, shlex.quote(str(self.config.path))]
        _command.extend(self.config.safe_arguments())
        completed_process: subprocess.CompletedProcess = (
            subprocess.run(  # nosec: Using shlex
                _command, stderr=subprocess.PIPE, timeout=10
            )
        )
        return VerificationStateData(
            rc=completed_process.returncode,
            type=self.__verification_type__,
            data=dict(stdout=completed_process.stdout, stderr=completed_process.stderr),
        )



#  Copyright 2021, Yahoo
#  Licensed under the terms of the Apache 2.0 license. See the LICENSE file in the project root for terms
import json
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Union
from uuid import UUID, uuid4

import yaml
from pydantic import Field

from ..utils.yaml import Dumper
from . import SchemaModel, SystemState
from .attack import AttackConfig
from .verification import VerificationConfig


class TestPlanSchema(SchemaModel):
    """
    Defines the schema for the Test Plan.
    """

    __test__ = False

    description: str = Field(
        default="", description="Description of the Test performed in this test plan"
    )

    verification: List[VerificationConfig] = Field(
        default=list(),
        description="List of verification configurations that will be executed to verify the system is in favorable condition",
    )

    attack: AttackConfig = Field(
        ..., description="The configuration that will be used to create chaos"
    )

    def filter_verification_by_state(
        self, system_state: SystemState
    ) -> List[VerificationConfig]:
        """
        Get all the verification configurations for a particular state
        Args:
            system_state: SystemState to filter the list

        Returns:
            Filtered list of verification configuration that are to be executed in a particular state
        """
        _filtered_config = list()
        for verification_config in self.verification:
            assert isinstance(verification_config.states, list)
            if system_state in verification_config.states:
                _filtered_config.append(verification_config)

        return _filtered_config

    class Config:
        @staticmethod
        def schema_extra(schema: Dict[str, Any], model) -> None:
            for prop in schema.get("properties", {}).values():
                prop.pop("title", None)

            schema_extra = {
                "$schema": "https://json-schema.org/draft-07/schema",
                "$id": "https://resilience.yahoo.com/schema.json",
            }
            schema.update(schema_extra)


class TestPlan(TestPlanSchema):
    """
    The Test Plan Dataclass.
    """

    __test__ = False

    id: UUID = Field(
        default_factory=uuid4,
        description="Schema for the test plan identifier",
        examples=[
            "061fc077-b95b-478b-87b6-73c29cb33c04",
            "6402e065-9c90-4719-b98c-ad03152e1238",
        ],
    )

    @classmethod
    def load_file(cls, path: Union[str, Path]) -> "TestPlan":
        path = Path(path)
        with open(path, "r") as file:
            data = yaml.safe_load(file)
        return cls(**data)

    def to_serialized_dict(
        self,
        *,
        include=None,
        exclude=None,
        by_alias: bool = False,
        skip_defaults: bool = None,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        encoder: Optional[Callable[[Any], Any]] = None,
        **dumps_kwargs: Any,
    ) -> dict:
        return json.loads(
            self.json(
                include=include,
                exclude=exclude,
                by_alias=by_alias,
                skip_defaults=skip_defaults,
                exclude_unset=exclude_unset,
                exclude_defaults=exclude_defaults,
                exclude_none=exclude_none,
                encoder=encoder,
                **dumps_kwargs,
            )
        )

    def export_to_file(
        self, path: Union[str, Path], *, yaml_format: bool = False, **kwargs
    ):
        path = Path(path)
        json_data = self.to_serialized_dict(**kwargs)
        with open(path, "w") as file:
            if yaml_format:
                yaml.dump(
                    json_data,
                    file,
                    default_flow_style=False,
                    sort_keys=False,
                    Dumper=Dumper,
                    **kwargs,
                )
            else:
                json.dump(json_data, file, indent=4)



#  Copyright 2021, Yahoo
#  Licensed under the terms of the Apache 2.0 license. See the LICENSE file in the project root for terms

import os
import warnings
from abc import ABC, abstractmethod
from datetime import datetime
from enum import IntEnum
from queue import LifoQueue, Queue
from threading import Thread
from types import SimpleNamespace
from typing import Any, Dict, Tuple

from pydantic import BaseModel, Field

from ..utils.builtins import BuiltinUtils
from .exceptions import AgentError


class AgentState(IntEnum):
    SKIPPED = -3
    ABORTED = -2
    ERROR = -1
    UNDEFINED = 0
    INIT = 1
    SETUP = 2
    RUNNING = 3
    COMPLETED = 4
    TEARDOWN = 5
    DONE = 6


class AgentPriority(IntEnum):
    VERY_HIGH_PRIORITY = 0
    HIGH_PRIORITY = 1
    MODERATE_PRIORITY = 2
    LOW_PRIORITY = 3
    VERY_LOW_PRIORITY = 4
    UNDEFINED_PRIORITY = -1


class AgentMonitoringDataPoint(BaseModel):
    timestamp: datetime = Field(default_factory=datetime.utcnow)
    data: Dict[str, Any] = Field(
        ..., description="Data from the agent at `timestamp` instant"
    )
    state: AgentState = Field(
        description="The state of the agent at `timestamp` instant"
    )


class AgentConfig(BaseModel):
    """
    The configuration of the agent that is to be run. This is the base configuration
    for any agent. Custom configuration attributes can be added to individual agent by inheriting
    this model.
    """

    name: str = Field(..., description="A one word identifier for the Agent.")
    description: str = Field(
        default="An awesome YChaos agent.",
        description="Multiline description of the agent in consideration",
    )
    priority: AgentPriority = Field(
        default=AgentPriority.UNDEFINED_PRIORITY,
        description="A priority assigned to the agent.",
    )

    # Runner configurations (Advanced configuration parameters)
    is_sudo: bool = Field(
        default=False,
        description="Setting this to true, requires the agent to run as root.",
    )

    # Be very careful when setting this key to False.
    # If you are aware of what you are doing, then go ahead.
    raise_on_state_mismatch: bool = Field(
        default=True, description="Raise error on state mismatch"
    )

    start_delay: int = Field(
        default=10, description="Give a delay of few seconds before running this agent"
    )

    def get_agent(self):
        """
        The Fallback factory method to use where the Agent
        instance depends on the configuration

        Returns:
            An agent subclass of Agent
        """
        pass


class TimedAgentConfig(AgentConfig):
    """
    The configuration of the agent which is constrained by a duration. This configuration
    is used for agents that are supposed to run for a particular amount of time before going
    to teardown state.
    """

    duration: int = Field(
        default=300,
        description="The duration for which this agent should run",
    )


class Agent(ABC):
    """
    An agent is an attack module that is configured to cause some kind of chaos on
    the target. A very simple of agent is the CPU Burn agent that is responsible
    for consuming CPU resources during the time interval its run.

    Each agent takes in one configuration object that is a subclass of AgentConfig.

    The agents have a lifecycle defined in `AgentState` each of them indicating
    the state in which the agent is currently in. The agent is advanced to the next state before
    executing any of the lifecycle methods.
    """

    def __init__(self, config):
        """
        Initialize an agent with a configuration

        Raises:
            InsufficientPermissionError: When the agent is configured to be sudo, but the agent is not run as root.

        Args:
            config: Agent configuration.
        """
        self.config = config

        self._runner = Thread(target=self.__run_exc_wrapper, name=config.name)
        self._stopper = Thread(
            target=self.__teardown_exc_wrapper, name=config.name + "_teardown"
        )
        self.stop_async_run: bool = False  # can be used as a flag to stop the attack and return from `run` method

        self.exception = Queue(-1)

        self._status = LifoQueue()
        self._state_history = list()

        self.preserved_state = SimpleNamespace(has_error=False, is_aborted=False)
        self.advance_state(AgentState.INIT)

    @abstractmethod
    def monitor(self) -> LifoQueue:  # pragma: no cover
        """
        Defines the implementation to monitor some stats for this agent and return a queue of
        the status

        Returns:
            A Queue of the status for this agent.
        """
        pass

    @abstractmethod
    def setup(self) -> None:
        """
        Defines the setup method. Used to set up the resources
        required for the agent. This is usually called before start

        Returns:
            None
        """
        self.advance_state(AgentState.SETUP)

    @abstractmethod
    def run(self) -> None:
        """
        Define what the agent does for the attack. Calling this method will block the calling thread.
        Caller has to take responsibility of updating the state of the agent once it is done.
        Use `start` or `start_async` for a better usage.

        Raises:
            AgentError: if `config.error_on_state_mismatch` is True and current state is not SETUP

        Returns:
            None
        """
        if self.current_state != AgentState.SETUP:
            if self.config.raise_on_state_mismatch:
                self.advance_state(AgentState.ABORTED)
                raise AgentError("Agent state is not in SETUP state. Bailing out")

            warnings.warn(
                "Agent is currently not in the SETUP state. Proceeding anyway"
            )

        if not self.is_runnable():
            raise AgentError("Agent not in executable state. Bailing out")
        self.advance_state(AgentState.RUNNING)

    def start(
        self,
        coro=BuiltinUtils.pass_coroutine,
        args: Tuple[Any, ...] = tuple(),
        interval=None,
    ) -> None:
        """
        A blocking start call. This method waits for the agent process to complete/exit.
        Calls a `coro` coroutine every `interval` seconds
        Also sets the agent state on COMPLETED/ERROR.
        Use `start_async()` method for the runner thread to run in background

        Args:
            coro: A function to call every `interval` seconds when the _runner is alive
            args: Arguments
            interval: Interval between 2 coroutine calls

        Returns:
            None
        """
        self._runner.start()
        while self._runner.is_alive():
            coro(*args)
            self._runner.join(interval)

        if self.exception.empty():
            self.advance_state(AgentState.COMPLETED)
        else:
            self.advance_state(AgentState.ERROR)

    def start_async(self) -> Thread:  # pragma: no cover
        """
        Unblocking call to start the run method. It is the responsibility of the
        caller to update the agent with its state, handle exceptions, etc.
        """
        self._runner.start()
        return self._runner

    def teardown_async(self) -> Thread:  # pragma: no cover
        """
        Non blocking call to start the teardown method. The caller takes full responsibility of handing

        Returns:
            teardown_async Thread instance
        """
        self._stopper.start()
        return self._stopper

    @abstractmethod
    def teardown(self) -> None:
        """
        This is called once the execute method is done to ensure a rollback is done
        to the state the agent entered.It wait for start_async thread to stop if its alive.

        Returns:
            None
        """
        self.advance_state(AgentState.TEARDOWN)
        self.stop_async_run = True
        if self._runner.is_alive():
            self._runner.join()

    def is_runnable(self) -> bool:
        """
        Fail Fast approach to find if the agent will be able to proceed to the next step.
        It is advised to run this method before running, start, run or start_async to
        avoid any mishaps in testing.

        Returns:
            True if able to proceed, False otherwise
        """
        if self.current_state < 0:
            return False
        if not self.exception.empty():
            return False
        if self.config.is_sudo and os.geteuid() != 0:
            return False

        return True

    def advance_state(self, state: AgentState):
        if self._state_history and self._state_history[-1] == state:
            return
        self._state_history.append(state)

    @property
    def current_state(self) -> AgentState:
        """
        Gets the current state of the agent.

        Returns:
            current state
        """
        if not self._state_history:
            # Never possible
            self._state_history = list()
            self.advance_state(AgentState.UNDEFINED)
        return self._state_history[-1]

    def __run_exc_wrapper(self):
        try:
            self.run()
        except Exception as e:
            self.exception.put(e)
            self.advance_state(AgentState.ERROR)

    def __teardown_exc_wrapper(self):
        try:
            self.teardown()
            self.advance_state(AgentState.DONE)
        except Exception as e:
            self.exception.put(e)
            self.advance_state(AgentState.ERROR)



#  Copyright 2021, Yahoo
#  Licensed under the terms of the Apache 2.0 license. See the LICENSE file in the project root for terms

import shutil
from pathlib import Path
from queue import LifoQueue
from stat import S_IREAD, S_IRGRP, S_IROTH
from tempfile import NamedTemporaryFile
from typing import List, Optional

from pydantic import Field, validate_arguments

from ..agent import Agent, TimedAgentConfig
from ..utils.annotations import log_agent_lifecycle


class TrafficBlockConfig(TimedAgentConfig):
    name = "traffic_block"
    description = "This agent modifies the the hosts /etc/hosts file to block traffic to certain hostnames"

    is_sudo = True

    hostsfile: Path = Field(
        description="The Filepath of hosts file", default=Path("/etc/hosts")
    )

    backup_hostsfile: Optional[Path] = Field(
        description="The filepath to store backup of hosts file. By default, the agent will create temporary file",
        default=None,
    )

    hosts: List[str] = Field(
        description="List of destination outbound hostnames to block",
        default=list(),
        examples=[["yahoo.com", "google.com"]],
    )


class TrafficBlock(Agent):
    LOCALHOST = "127.0.0.1"
    permission = int()

    @validate_arguments
    def __init__(self, config: TrafficBlockConfig):
        super(TrafficBlock, self).__init__(config)

        if config.backup_hostsfile is None:
            config.backup_hostsfile = Path(
                NamedTemporaryFile(mode="w+", delete=False).name
            )

    def monitor(self) -> LifoQueue:
        super(TrafficBlock, self).monitor()
        return self._status

    @log_agent_lifecycle
    def setup(self) -> None:
        super(TrafficBlock, self).setup()

        shutil.copy(self.config.hostsfile, self.config.backup_hostsfile)

        # Makes the Backup File as Read Only. This is a protection from
        # backup file getting deleted/modified accidentally. The file can still
        # be modified with root access.
        self.permission = Path(self.config.backup_hostsfile).lstat().st_mode
        Path(self.config.backup_hostsfile).chmod(S_IREAD | S_IRGRP | S_IROTH)

    @log_agent_lifecycle
    def run(self) -> None:
        super(TrafficBlock, self).run()
        with open(self.config.hostsfile, "a") as hosts_file:
            hosts_file.write("\n")
            for host in self.config.hosts:
                hosts_file.write(f"{self.LOCALHOST}\t{host}\n")

    @log_agent_lifecycle
    def teardown(self) -> None:
        super(TrafficBlock, self).teardown()

        # Restore the permissions for the file and move it to original location
        self.config.backup_hostsfile.chmod(self.permission)
        shutil.copy(self.config.backup_hostsfile, self.config.hostsfile)

        # Delete the backup hostsfile
        self.config.backup_hostsfile.unlink()



#  Copyright 2021, Yahoo
#  Licensed under the terms of the Apache 2.0 license. See the LICENSE file in the project root for terms
from abc import ABC, abstractmethod

import requests
from requests import Session

from ....app_logger import AppLogger
from ..data import VerificationData, VerificationStateData


class BaseVerificationPlugin:

    __verification_type__: str

    def __init__(
        self,
        config,
        state_data: VerificationData = VerificationData.parse_obj(dict()),
        **kwargs
    ):
        self.config = config
        self.state_data = state_data
        self.logger = AppLogger.get_logger(self.__class__.__name__)

        self.logger.bind(event="verification", type=self.__verification_type__)

    @abstractmethod
    def run_verification(self) -> VerificationStateData:
        pass  # Abstract method


class RequestVerificationPlugin(BaseVerificationPlugin, ABC):
    def _build_session(self):
        session = Session()

        session.params = self.config.params
        session.headers = self.config.headers

        if not self.config.verify:
            # Disable Insecure Request Warning if verify=False (User knows what he is doing)
            from requests.packages.urllib3.exceptions import (
                InsecureRequestWarning,
            )

            requests.packages.urllib3.disable_warnings(InsecureRequestWarning)

        session.verify = self.config.verify

        if self.config.basic_auth:
            session.auth = (
                self.config.basic_auth[0],
                self.config.basic_auth[1].get_secret_value(),
            )

        if self.config.bearer_token:
            session.headers["Authorization"] = (
                "Bearer " + self.config.bearer_token.get_secret_value()
            )

        session.cert = self.config.cert

        return session

    @abstractmethod
    def run_verification(self) -> VerificationStateData:
        pass  # Abstract method



#  Copyright 2021, Yahoo
#  Licensed under the terms of the Apache 2.0 license. See the LICENSE file in the project root for terms

import re
from enum import Enum, EnumMeta
from types import DynamicClassAttribute, SimpleNamespace
from typing import Any, Iterable, List, Optional, Type, TypeVar

T = TypeVar("T")


class BuiltinUtils:
    class Request:
        HTTP_METHODS = ("GET", "POST", "HEAD", "PATCH", "DELETE")

        @classmethod
        def validate_method(cls, v):
            if v in cls.HTTP_METHODS:
                return v
            else:
                raise ValueError("Unknown HTTP method")

    class Float:
        NAN = float("NaN")

        @classmethod
        def parse(cls, val: Any, default: float) -> float:
            try:
                return float(val)
            except Exception:
                return default

    @classmethod
    def raise_error(cls, e):
        raise e

    @classmethod
    def return_if_true(cls, return_value: Any, comparison: Any, default: Any = None):
        """
        Idiomatic in writing the below logic

        ```python
        return return_value if comparison else default
        ```

        Args:
            return_value: The return value
            comparison: Comparison logic
            default: Default value to return

        Returns:
            return_value if comparison is True else default
        """
        return return_value if comparison else default

    @classmethod
    def wrap_if_non_iterable(cls, obj: Any):
        """
        Wraps an object into a List only if the object is not
        an iterable. If the object is already an Iterable, the method returns
        the object type converted to List.

        Args:
            obj: Any object

        Returns:
            Wrapped list if non iterable
        """
        if isinstance(obj, Iterable):
            return list(obj)
        else:
            return cls.wrap_as_list(obj)

    @classmethod
    def wrap_as_list(cls, obj) -> List:
        """
        Wrap an object to a List.
        Args:
            obj:

        Returns:
            Wrapped list
        """
        return [
            obj,
        ]

    @classmethod
    def pass_coroutine(cls, *args, **kwargs):
        """This method literally does nothing"""
        pass


class AEnum(Enum, metaclass=EnumMeta):
    """
    Advanced Enumeration to add a metadata to each of the Enum object.
    This will add a 2 level mapping for NAME -> VALUE -> METADATA. The label
    is optional and can be set to a simple

    Special metadata attributes
    1. __aliases__ : (tuple) Defines serializable aliases for the same Enum
    2. __desc__ : (str) Defines the documentation for an enum member

    """

    def __new__(cls: Type[T], value, metadata: Optional[SimpleNamespace] = None):
        obj = object.__new__(cls)
        obj._value_ = value
        obj.metadata = metadata

        # Add Aliases to Serializer Dictionary
        for alias in getattr(metadata, "__aliases__", tuple()):
            cls._value2member_map_[alias] = obj  # type: ignore

        obj.__doc__ = getattr(metadata, "__desc__", None)
        return obj

    @DynamicClassAttribute
    def value(self) -> str:
        # mypy causes issues without this
        return self._value_


class FQDN(str):
    _regex = r"^((?![-])[-A-Z\d]{1,63}(?<!-)[.])*(?!-)[-A-Z\d]{1,63}(?<!-)[.]?$"

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, fqdn: str):
        if len(fqdn) > 255:
            raise ValueError(f"{fqdn} is not a valid FQDN")
        fqdn = fqdn[:-1] if fqdn[-1] == "." else fqdn
        allowed = re.compile(cls._regex, re.IGNORECASE)
        if all(allowed.match(x) for x in fqdn.split(".")):
            return fqdn
        else:
            raise ValueError(f"{fqdn} is not a valid FQDN")

    def __new__(cls, *args, **kwargs):
        return cls.validate(args[0])



#  Copyright 2021, Yahoo
#  Licensed under the terms of the Apache 2.0 license. See the LICENSE file in the project root for terms

from argparse import ArgumentParser

from ...utils.argparse import SubCommandParsersAction
from .. import YChaosSubCommand
from .attack import Attack

__all__ = ["Agent"]


class Agent(YChaosSubCommand):
    """
    The Agent subcommand used to operate on YChaos agents. Most of the subcommands
    under `agent` requires testplan.
    """

    name = "agent"
    help = "The agent subcommand of YChaos"

    @classmethod
    def build_parser(cls, parser: ArgumentParser) -> ArgumentParser:
        test_plan_command_subparser = parser.add_subparsers(
            action=SubCommandParsersAction, dest=f"_cmd.{cls.name}"
        )
        assert isinstance(test_plan_command_subparser, SubCommandParsersAction)

        test_plan_command_subparser.add_parser(cls=Attack)
        return parser



#  Copyright 2021, Yahoo
#  Licensed under the terms of the Apache 2.0 license. See the LICENSE file in the project root for terms
import os
import shlex
import subprocess  # nosec
from pathlib import Path
from typing import Union

from pydantic import validate_arguments


class SysCtl:
    """
    Provides a Utility class to operate with sysctl command.
    """

    ROOT_PATH = Path("/proc/sys")

    _cmd = "sysctl"
    _sudo = "sudo"

    @classmethod
    def get(cls, variable: str) -> bytes:
        """
        Get a sysctl variable's value as bytes.
        Args:
            variable: Variable name. Can contain `.`

        Raises:
            PermissionError: If the sysctl variable not visible to the current user

        Returns:
            A bytes object that contains the value
        """
        sysctl_var_path = cls.ROOT_PATH.joinpath(variable.replace(".", "/"))
        cls.is_variable(variable, raise_error=True)
        return sysctl_var_path.read_bytes()

    @classmethod
    @validate_arguments
    def set(cls, variable: str, value: Union[str, bytes]) -> bool:
        """
        Set a sysctl variable with a custom value. The value and the variable are shell escaped to
        avoid shell injection.
        Args:
            variable: Variable name. This can contain `.`
            value: value for the variable.

        Returns:
            True if the method was able to set the variable, False otherwise
        """
        # Protect against shell injection vuln using shlex.quote
        if isinstance(value, bytes):
            value = str(value, "UTF-8")
        _variable, _value = shlex.quote(variable), shlex.quote(value)
        _cmd = [cls._cmd, _variable, _value]
        if os.geteuid() != 0:
            _cmd.insert(0, cls._sudo)
        proc = subprocess.run(  # nosec
            _cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        if proc.returncode == 0:
            return True
        return False

    @classmethod
    def is_variable(cls, key: str, raise_error: bool = False) -> bool:
        """
        Determine if a key is a valid sysctl variable.
        Args:
            key:
            raise_error: Raise `KeyError` if not a variable

        Raises:
            KeyError: if `raise_error` is True and `key` is not a valid sysctl variable.

        Returns:
            True if valid sysctl variable, False otherwise
        """
        sysctl_var_path = cls.ROOT_PATH.joinpath(key.replace(".", "/"))
        if not sysctl_var_path.is_file():
            if raise_error:
                raise KeyError(
                    "Cannot find {sysctl_var_path}. No such sysctl variable.".format(
                        sysctl_var_path=sysctl_var_path
                    )
                )
            else:
                return False
        return True



#  Copyright 2021, Yahoo
#  Licensed under the terms of the Apache 2.0 license. See the LICENSE file in the project root for terms

#  Copyright 2021, Yahoo
#  Licensed under the terms of the Apache 2.0 license. See the LICENSE file in the project root for terms
from typing import Optional

from pydantic import BaseModel, validate_arguments

from ....testplan.verification import HTTPRequestVerification
from ..data import VerificationData, VerificationStateData
from .BaseVerificationPlugin import RequestVerificationPlugin


class HTTPRequestVerificationPlugin(RequestVerificationPlugin):
    class HttpVerificationData(BaseModel):
        url: str  # Making this string since this is built with only trusted data
        status_code: Optional[int]
        latency: Optional[int]
        error: Optional[str]
        error_desc: Optional[str]

    __verification_type__ = "http_request"

    @validate_arguments
    def __init__(
        self,
        config: HTTPRequestVerification,
        state_data: VerificationData = VerificationData.parse_obj(dict()),
    ):
        super(HTTPRequestVerificationPlugin, self).__init__(config, state_data)
        self._session = self._build_session()

    def run_verification(self) -> VerificationStateData:
        _rc = 0
        _data = list()

        for _ in range(self.config.count):
            counter_data = list()

            for url in self.config.urls:
                try:
                    response = self._session.request(
                        self.config.method,
                        url=str(url),
                        timeout=self.config.timeout / 1000,
                    )
                    if (
                        response.status_code not in self.config.status_codes
                        or (response.elapsed.microseconds / 1000) > self.config.latency
                    ):
                        _rc = 1
                        counter_data.append(
                            self.HttpVerificationData(
                                url=response.url,
                                status_code=response.status_code,
                                latency=response.elapsed.microseconds / 1000,
                            ).dict()
                        )
                except Exception as e:
                    _rc = 1
                    counter_data.append(
                        self.HttpVerificationData(
                            url=str(url), error=e.__class__.__name__, error_desc=str(e)
                        ).dict()
                    )
            _data.append(counter_data)

        return VerificationStateData(
            rc=_rc, type=self.__verification_type__, data=_data
        )



#  Copyright 2021, Yahoo
#  Licensed under the terms of the Apache 2.0 license. See the LICENSE file in the project root for terms
import os
from types import SimpleNamespace
from typing import Any

from pydantic import Field

from ..utils.builtins import AEnum
from . import SchemaModel


class SecretType(AEnum):
    """
    Defines the type of secret the caller wants to fetch

    Attributes:
        env: The secret is to be fetched from Environment variable.

    """

    ENV = "env", SimpleNamespace(parser=lambda v: os.getenv(v))
    # Defines the secret to be fetched from environment variable


class Secret(SchemaModel):
    """
    Secret object defines the ways of fetching a secret to populate a value. For example, the login
    password to a host. The type of secret defines what kind of secret, the tool is trying to fetch
    and id is the unique identifier for that secret.

    A simple example is a Environment Variable. If the type=`env` and id=`HOST_PASSWORD`, then the tool
    will read the environment variable to determine the secret to use. This will allow the users
    to not reveal the secrets in testplan.
    """

    type: SecretType = Field(
        default=SecretType.ENV, description="defines the type of secret to be fetched. "
    )
    id: Any = Field(
        ...,
        description="The public identifier data which can be used to fetch the secret",
    )

    def get_secret_value(self):
        """
        Return the secret value.
        Returns:
            the secret value fetched from the parser.
        """
        return self.type.metadata.parser(self.id)



#  Copyright 2021, Yahoo
#  Licensed under the terms of the Apache 2.0 license. See the LICENSE file in the project root for terms
from abc import ABC
from argparse import ArgumentParser, Namespace
from pathlib import Path
from typing import Any, Optional, Union

from pydantic import validate_arguments

from ..core.verification.controller import VerificationController
from ..core.verification.data import VerificationStateData
from ..testplan import SystemState
from ..testplan.verification import VerificationConfig, VerificationType
from . import YChaosCLIHook, YChaosTestplanInputSubCommand

__all__ = ["Verify"]


class Verify(YChaosTestplanInputSubCommand):
    """
    The `verify` subcommand of YChaos is used to verify the state of the system. This
    subcommand requires a valid testplan which can be provided with the -t/--testplan argument.
    The subcommand also requires a valid state at which the system is verified.
    """

    name = "verify"
    help = "The verification subcommand of YChaos"

    def __init__(self, **kwargs):
        super(Verify, self).__init__(**kwargs)

        self.test_plan_path: Path = kwargs.pop("testplan")
        self.state: SystemState = SystemState(kwargs.pop("state", None).upper())

        self.dump_yaml: Optional[Path] = kwargs.pop("dump_yaml", None)
        self.dump_json: Optional[Path] = kwargs.pop("dump_json", None)

        self.state_data_path: Optional[Path] = kwargs.pop("state_data", None)

    @classmethod
    def build_parser(cls, parser: ArgumentParser) -> ArgumentParser:
        parser = super(Verify, cls).build_parser(parser)
        parser.add_argument(
            "-s",
            "--state",
            choices=[x.value.lower() for x in list(SystemState)],
            help="System state to verify",
            default="steady",
            metavar="state",
        )

        report_argument_group = parser.add_argument_group("verification reports")
        report_argument_group.add_argument(
            "--dump-yaml",
            type=Path,
            help="Store the verification data in YAML format",
            required=False,
            metavar="path",
        )

        report_argument_group.add_argument(
            "--dump-json",
            type=Path,
            help="Store the verification data in JSON format",
            required=False,
            metavar="path",
        )

        parser.add_argument(
            "--state-data",
            type=Path,
            help="The path of the verification data state file (JSON/YAML)",
            required=False,
            metavar="path",
        )

        return parser

    def get_state_data(self):
        self.console.log("Getting state data")
        self.console.line()
        try:
            import yaml

            path = Path(self.state_data_path)
            with open(path, "r") as file:
                state_data = yaml.safe_load(file)
            return state_data
        except IsADirectoryError as is_directory:
            self.set_exitcode(1)
            self.console.print(
                f":file_folder: The input path ({self.state_data_path}) is not a valid state data file",
                style="indian_red",
            )
        except FileNotFoundError as file_not_found_error:
            self.set_exitcode(1)
            self.console.print(
                ":mag: {file} [italic]not found[/italic]".format(
                    file=str(self.state_data_path)
                ),
                style="indian_red",
            )
        return None

    @validate_arguments(config=dict(arbitrary_types_allowed=True))
    def _generate_verification_report(
        self,
        verification_controller: VerificationController,
        output_format: str,
        report_file_path: Union[Path, str],
    ):
        self.console.log(
            f"Writing {self.state.value.lower()} state verification data to {report_file_path}, format={output_format}"
        )
        report_output_base_path = Path(report_file_path).parent

        try:
            if not report_output_base_path.is_dir():
                report_output_base_path.mkdir(parents=True, exist_ok=True)
        except PermissionError as permission_error:
            self.console.log(
                ":file_folder: [italic]Permission denied to create report directory[/italic]",
                style="indian_red",
            )
            return

        try:
            with open(report_file_path, "w") as fp:
                verification_controller.dump_verification(
                    fp, output_format=output_format
                )
        except PermissionError as permission_error:
            self.console.log(
                ":file_folder: [italic]Permission denied to create report file[/italic]",
                style="indian_red",
            )

    def verify_system_state(self):
        testplan = self.get_validated_test_plan(self.test_plan_path)

        if self._exitcode != 0:
            return

        state_data = list()
        if self.state_data_path:
            state_data = self.get_state_data()
            if self._exitcode != 0:
                return

        # section Hooks
        class VerificationHook(YChaosCLIHook, ABC):
            def __init__(self, app, state: SystemState):
                super(VerificationHook, self).__init__(app)
                self.state = state

        class OnEachPluginStartHook(VerificationHook):
            def __call__(self, index: int, config: VerificationConfig):
                self.console.log(
                    f"Running [i]{self.state.value.lower()}[/i] state verification of type={config.type.value}[{index}]"
                )

        class OnPluginNotFoundHook(VerificationHook):
            def __call__(self, index: int, plugin_type: VerificationType):
                self.console.log(
                    f"The verification plugin type=[i]{plugin_type.value}[/i][{index}] is not available for use."
                )

        class OnEachPluginEndHook(VerificationHook):
            def __call__(
                self,
                index: int,
                config: VerificationConfig,
                verified_state_data: VerificationStateData,
            ):
                self.console.log(
                    (
                        f"Completed [i]{self.state.value.lower()}[/i] state verification of type={config.type.value};"
                        f" verified={verified_state_data.rc==0}"
                    )
                )

        # end section

        verification_controller = VerificationController(
            testplan, self.state, state_data
        )
        verification_controller.register_hook(
            "on_each_plugin_start", OnEachPluginStartHook(self.app, self.state)
        )
        verification_controller.register_hook(
            "on_each_plugin_end", OnEachPluginEndHook(self.app, self.state)
        )
        verification_controller.register_hook(
            "on_plugin_not_found", OnPluginNotFoundHook(self.app, self.state)
        )

        self.console.log(
            f"Starting [i]{self.state.value.lower()}[/i] state verification."
        )
        is_verified = verification_controller.execute()

        self.set_exitcode(int(not is_verified))

        self.console.line()
        if is_verified:
            self.console.print(
                f"The system is verified to be in {self.state.value.lower()} state",
                style="green",
            )
        else:
            self.console.print(
                f"The system is not verified to be in {self.state.value.lower()} state",
                style="red",
            )

        self.console.line()
        if self.dump_json:
            self._generate_verification_report(
                verification_controller,
                report_file_path=self.dump_json,
                output_format="json",
            )

        if self.dump_yaml:
            self._generate_verification_report(
                verification_controller,
                report_file_path=self.dump_yaml,
                output_format="yaml",
            )

    @classmethod
    def main(cls, args: Namespace) -> Any:  # pragma: no cover
        verification_command = Verify(**vars(args))

        verification_command.verify_system_state()

        return verification_command._exitcode



#  Copyright 2021, Yahoo
#  Licensed under the terms of the Apache 2.0 license. See the LICENSE file in the project root for terms

import math
import random
from datetime import datetime
from types import SimpleNamespace
from typing import Dict, List, Tuple, Union

from pydantic import Field, validate_arguments, validator

from ....utils.builtins import AEnum, BuiltinUtils
from ... import SchemaModel, SystemState


class TimeSeriesDataAggregator:
    """
    The class containing the implementations of different Time Series data Aggregator
    """

    @classmethod
    def get_filtered_data(cls, data: Dict[datetime, float]) -> Dict[datetime, float]:
        """
        Filter the time series data by removing NAN data
        Args:
            data: Time series data

        Returns:
            Filtered data
        """
        return dict([(_k, _v) for _k, _v in data.items() if not math.isnan(_v)])

    @classmethod
    def avg(cls, data: Dict[datetime, float]) -> float:
        """
        Returns the Average value of a time series data filtering out NAN
        Args:
            data: Time series data

        Returns:
            Average value over the time
        """
        _filtered_data = cls.get_filtered_data(data)
        return BuiltinUtils.return_if_true(
            sum(_filtered_data.values()) / len(_filtered_data),
            _filtered_data,
            BuiltinUtils.Float.NAN,
        )

    @classmethod
    def latest(cls, data: Dict[datetime, float]) -> float:
        """
        Returns the latest value of a time series data filtering out NAN
        Args:
            data: Time series data

        Returns:
            Latest value over the time
        """
        _filtered_data = cls.get_filtered_data(data)
        return BuiltinUtils.return_if_true(
            _filtered_data[max(_filtered_data)], _filtered_data, BuiltinUtils.Float.NAN
        )

    @classmethod
    def oldest(cls, data: Dict[datetime, float]) -> float:
        """
        Returns the oldest value of a time series data filtering out NAN
        Args:
            data: Time series data

        Returns:
            Latest value over the time
        """
        _filtered_data = cls.get_filtered_data(data)
        return BuiltinUtils.return_if_true(
            _filtered_data[min(_filtered_data)], _filtered_data, BuiltinUtils.Float.NAN
        )

    @classmethod
    def max(cls, data: Dict[datetime, float]) -> float:
        """
        Returns the max value of a time series data filtering out NAN
        Args:
            data: Time series data

        Returns:
            Max value
        """
        _filtered_data = cls.get_filtered_data(data)
        return BuiltinUtils.return_if_true(
            max(_filtered_data.values()), _filtered_data, BuiltinUtils.Float.NAN
        )

    @classmethod
    def min(cls, data: Dict[datetime, float]) -> float:
        """
        Returns the min value of a time series data filtering out NAN
        Args:
            data: Time series data

        Returns:
            Min value
        """
        _filtered_data = cls.get_filtered_data(data)
        return BuiltinUtils.return_if_true(
            min(_filtered_data.values()), _filtered_data, BuiltinUtils.Float.NAN
        )

    @classmethod
    def random(cls, data: Dict[datetime, float]) -> float:
        """
        Returns the random value of a time series data filtering out NAN
        Args:
            data: Time series data

        Returns:
            Random value
        """
        _filtered_data = cls.get_filtered_data(data)
        return BuiltinUtils.return_if_true(
            random.choice(  # nosec : Not using for Crypto purpose
                list(_filtered_data.values())
            ),
            _filtered_data,
            BuiltinUtils.Float.NAN,
        )


class MetricsAggregator(AEnum):
    """
    The Metrics aggregator options. Allows the user to transform a
    time series data into comparable data.
    """

    AVG = "avg", SimpleNamespace(
        aggregate=TimeSeriesDataAggregator.avg,
        __desc__="Gets the Average of all valid datapoints",
    )

    LATEST = "latest", SimpleNamespace(
        aggregate=TimeSeriesDataAggregator.latest,
        __desc__="Gets the latest valid datapoint",
    )

    OLDEST = "oldest", SimpleNamespace(
        aggregate=TimeSeriesDataAggregator.oldest,
        __desc__="Gets the oldest valid datapoint",
    )

    RANDOM = "random", SimpleNamespace(
        aggregate=TimeSeriesDataAggregator.random,
        __desc__="Gets a random valid datapoint",
    )

    MAX = "max", SimpleNamespace(
        aggregate=TimeSeriesDataAggregator.max,
        __desc__="Gets the largest valid datapoint",
    )

    MIN = "min", SimpleNamespace(
        aggregate=TimeSeriesDataAggregator.min,
        __desc__="Gets the smallest valid datapoint",
    )

    SLOPE = "slope", SimpleNamespace(
        aggregate=lambda data: BuiltinUtils.raise_error(
            NotImplementedError("This aggregator is not implemented")
        ),
        __desc__="Gets the slope of the datapoints",
    )


class MetricsComparator(AEnum):

    LT = "lt", SimpleNamespace(
        __aliases__=("<",),
        compare=lambda data, expected: data < expected,
        relational=True,
    )

    LE = "le", SimpleNamespace(
        __aliases__=("<=",),
        compare=lambda data, expected: data <= expected,
        relational=True,
    )

    GT = "gt", SimpleNamespace(
        __aliases__=(">",),
        compare=lambda data, expected: data > expected,
        relational=True,
    )

    GE = "ge", SimpleNamespace(
        __aliases__=(">=",),
        compare=lambda data, expected: data >= expected,
        relational=True,
    )

    EQ = "eq", SimpleNamespace(
        __aliases__=("==",),
        compare=lambda data, expected: data == expected,
        relational=True,
    )

    NEQ = "neq", SimpleNamespace(
        __aliases__=("!=",),
        compare=lambda data, expected: data != expected,
        relational=True,
    )

    RANGE = "range", SimpleNamespace(
        __aliases__=("()", "(]", "[)", "[]"),
        compare=lambda range_type, data, expected_range: MetricsComparator.range_compare(  # type: ignore
            range_type, data, expected_range
        ),
        relational=False,
    )

    # The below comparators can only be used for state bound criteria

    PCT = "pct", SimpleNamespace(
        __aliases__=("%%",),
        compare=lambda new_data, old_data, expected: MetricsComparator.pct_compare(  # type: ignore
            new_data, old_data, expected
        ),
        relational=False,
    )

    @classmethod
    @validate_arguments
    def pct_compare(
        cls, new_data: float, old_data: float, expected: Union[float, Tuple]
    ) -> bool:
        """
        Calculate the percentage variation from new_val and old_val

        Args:
            new_data: New Value from data
            old_data: Old value from saved data
            expected: Expected percentage change

        Returns:
            True if the condition meets
        """
        pct_change = ((new_data - old_data) / old_data) * 100
        if isinstance(expected, tuple):
            return (
                BuiltinUtils.Float.parse(expected[0], -math.inf)
                <= pct_change
                <= BuiltinUtils.Float.parse(expected[1], math.inf)
            )
        else:
            return pct_change == expected

    @classmethod
    @validate_arguments
    def range_compare(
        cls,
        range_type: str,
        data: float,
        expected_range: Tuple,
    ) -> bool:
        """

        Args:
            range_type: The range type that depicts inclusiveness
            data: Data to be compared
            expected_range: Expected range for the data to be within

        Returns:

        """
        if range_type == MetricsComparator.RANGE.value:
            range_type = "()"

        if range_type == "()":
            return (
                BuiltinUtils.Float.parse(expected_range[0], -math.inf)
                < data
                < BuiltinUtils.Float.parse(expected_range[1], math.inf)
            )
        elif range_type == "[)":
            return (
                BuiltinUtils.Float.parse(expected_range[0], -math.inf)
                <= data
                < BuiltinUtils.Float.parse(expected_range[1], math.inf)
            )
        elif range_type == "(]":
            return (
                BuiltinUtils.Float.parse(expected_range[0], -math.inf)
                < data
                <= BuiltinUtils.Float.parse(expected_range[1], math.inf)
            )
        else:
            return (
                BuiltinUtils.Float.parse(expected_range[0], -math.inf)
                <= data
                <= BuiltinUtils.Float.parse(expected_range[1], math.inf)
            )


class ComparisonCondition(SchemaModel):

    comparator: MetricsComparator = Field(
        ...,
        description="Comparison condition to compare between the metrics data and fetched value",
    )

    value: Union[float, Tuple] = Field(
        ..., description="Numerical value/range to be used for comparison"
    )

    # Resolve Aliases
    @validator("comparator", pre=True)
    def resolve_comparator(cls, v, values):
        values["_comparator"] = v  # Store the actual value in a private attribute
        return v


class MultipleConditionalsMetricsVerificationCriteria(SchemaModel):
    """
    Defines the Metrics Verification criteria
    """

    aggeragator: MetricsAggregator = Field(
        default=MetricsAggregator.AVG, description="Data aggregator"
    )

    conditionals: List[ComparisonCondition] = Field(
        default=list(),
        description="The conditionals out of which any one needs to pass for the criteria to be marked as passed",
    )


class StateBoundMetricsVerificationCriteria(SchemaModel):

    aggeragator: MetricsAggregator = Field(
        default=MetricsAggregator.AVG, description="Data aggregator"
    )

    criteria: Dict[SystemState, ComparisonCondition] = Field(
        ...,
        description="Metrics verification criteria with state.",
    )

    def get_criteria(self, state: SystemState):
        return self.criteria[state.value]  # type: ignore

    class Config:
        use_enum_values = True  # To make the model JSON serializable



#  Copyright 2021, Yahoo
#  Licensed under the terms of the Apache 2.0 license. See the LICENSE file in the project root for terms
import os
import sys
from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser, Namespace
from collections import OrderedDict
from pathlib import Path
from typing import Any, Dict, Iterable, List, Optional, Union

from rich.console import Console
from rich.table import Table

from ..app_logger import AppLogger
from ..settings import ApplicationSettings, DevSettings, ProdSettings, Settings
from ..utils.argparse import SubCommandParsersAction
from . import YChaosArgumentParser, YChaosSubCommand
from .agent.main import Agent
from .execute import Execute
from .manual import Manual
from .testplan import TestPlan
from .verify import Verify


class YChaos:
    """
    The YChaos CLI class. The `main` method of this class is the starting point of
    the CLI which takes in the program arguments. See YChaos CLI documentation for
    more details.

    To see the usage of the YChaos CLI, run `ychaos -h` on the terminal
    """

    settings = ApplicationSettings.get_instance()

    @classmethod
    def main(cls, program_arguments: list) -> None:
        """
        All good things begin some place, and that's here.
        Args:
            program_arguments: List of Program Arguments

        Returns:
            None
        """
        ychaos_cli = YChaosArgumentParser(
            prog=cls.settings.PROG,
            formatter_class=ArgumentDefaultsHelpFormatter,
            __root__=YChaosRoot,
        )

        # YChaos CLI version
        ychaos_cli.add_argument(
            "-v",
            "--version",
            action="version",
            version=f"v{cls.settings.get_version()} [yahoo/ychaos]",
        )

        # Verbosity Argument Group
        verbosity_argument_group = ychaos_cli.add_argument_group("verbosity")
        verbosity_argument_group.add_argument(
            "-V",
            "--verbose",
            action="count",
            help="Increase verbosity of logs (INFO)",
            default=0,
            required=False,
        )
        verbosity_argument_group.add_argument(
            "--debug",
            action="store_true",
            help="Enable debug mode",
            default=False,
            required=False,
        )

        # Configuration for the ychaos Command Line Interface
        ychaos_cli.add_argument(
            "-c",
            "--config",
            choices=["dev", "prod"],
            default="prod",
            help="Set YChaos CLI configuration (prod)",
            metavar="config",
        )

        ychaos_cli.add_argument(
            "--no-color",
            action="store_true",
            default=bool(os.getenv("NO_COLOR", False)),
            help="Disable color on console output ($NO_COLOR)",
        )

        # Arguments for creating HTML & Text Reports
        report_argument_group = ychaos_cli.add_argument_group("reports")
        report_argument_group.add_argument(
            "--text-report",
            type=Path,
            default=None,
            required=False,
            help="Generate a text report from the YChaos execution",
            metavar="path",
        )
        report_argument_group.add_argument(
            "--html-report",
            type=Path,
            default=None,
            required=False,
            help="Generate a HTML report from YChaos execution",
            metavar="path",
        )
        report_argument_group.add_argument(
            "--log-file",
            type=Path,
            default=os.getenv("YCHAOS_LOG_FILE"),
            required=False,
            help=("The file to store application logs. ($YCHAOS_LOG_FILE)"),
            metavar="path",
        )

        ychaos_cli_subparsers = ychaos_cli.add_subparsers(
            action=SubCommandParsersAction,
            dest=cls.settings.COMMAND_IDENTIFIER.format(cls.settings.PROG),
        )

        assert isinstance(ychaos_cli_subparsers, SubCommandParsersAction)

        # Subcommands
        ychaos_cli_subparsers.add_parser(cls=TestPlan)
        ychaos_cli_subparsers.add_parser(cls=Manual)
        ychaos_cli_subparsers.add_parser(cls=Agent)
        ychaos_cli_subparsers.add_parser(cls=Verify)
        ychaos_cli_subparsers.add_parser(cls=Execute)

        args = ychaos_cli.parse_args(program_arguments)

        args.app = App(args, ychaos_cli)

        # Start the Application
        args.app.start()

        # Call the right method for the subcommand
        try:
            exitcode = ychaos_cli.run_command(args)
        except Exception as unknown_error:  # pragma: no cover
            exitcode = 255
            args.app.unknown_error()

        # Teardown
        args.app.teardown(exitcode)

        sys.exit(exitcode)


class App:
    def __init__(self, args: Namespace, cli: Optional[ArgumentParser] = None):
        Settings(args.config)

        args.verbose = min(2, args.verbose)

        self.args = args
        self.console = Console(record=True, no_color=args.no_color)
        self.settings: Union[DevSettings, ProdSettings] = Settings.get_instance()

        self.cli = cli

        if args.log_file:
            self.settings.LOG_FILE_PATH = args.log_file

        AppLogger()

    def start(self) -> None:
        AppLogger.start()
        self.console.clear()
        self.console.rule(
            title=self.settings.APP_DESC,
            style="magenta",
        )

        if self.args.cls != YChaosRoot:
            self.print_cli_configuration()
            self.console.line()
            self.console.log("Starting app")

    def get_command_tree(self) -> List[str]:
        """
        Determines the command tree of the invoked command.

        Returns:
            List of commands starting with `ychaos` and ending at the sub-command invoked
        """
        _args = vars(self.args)

        parent = self.settings.PROG
        branch = _args.get(self.settings.COMMAND_IDENTIFIER.format(parent), None)

        tree = []
        while branch is not None:
            tree.append(parent)
            parent = branch
            branch = _args.get(self.settings.COMMAND_IDENTIFIER.format(parent), None)

        tree.append(parent)
        return tree

    def _walk_parser(self, parser: ArgumentParser, tree):
        """
        Walk the Argument parser and collect all the subcommands and its help message.

        Args:
            parser: Argument Parser

        Returns:
            Tree dictionary
        """
        if parser._subparsers is not None:
            for action in parser._subparsers._actions:
                if isinstance(action, SubCommandParsersAction):
                    for command, subparser in action.choices.items():
                        tree[subparser.prog] = subparser.format_help()
                        tree = self._walk_parser(subparser, tree)
        return tree

    def manual_entry(self) -> Dict[str, str]:
        """
        Determines the manual entry of the YChaos CLI and returns a program
        aware dictionary of cli -> usage mapping

        Returns:
            Dictionary of command to help message mapping
        """
        _tree = OrderedDict()
        assert self.cli is not None
        _tree[self.settings.PROG] = self.cli.format_help()
        self._walk_parser(self.cli, _tree)

        return _tree

    def is_debug_mode(self) -> bool:
        """
        Returns if the app was initialized with debug mode
        Returns:
            True if debug mode
        """
        return self.args.debug

    def print_cli_configuration(self):
        self.console.line()
        table = Table(title="YChaos CLI configuration", header_style="bold green")
        table.add_column("Configuration", style="bold sea_green2")
        table.add_column("Value")

        _renderable_config = vars(self.args).copy()
        _renderable_config.pop("app")
        _renderable_config.pop("cls")

        _renderable_config["_command_"] = " :arrow_right: ".join(
            self.get_command_tree()
        )

        for k, v in sorted(_renderable_config.items()):
            if not k.startswith("_cmd"):
                if isinstance(v, Iterable) and not isinstance(v, str):
                    v = "\n".join([str(_v) for _v in v])
                if v is not None:
                    table.add_row(str(k), str(v))

        self.console.print(table)

    def unknown_error(self):
        self.console.line()
        self.console.print_exception(extra_lines=2)

    def teardown(self, exitcode: int) -> None:
        if self.args.cls != YChaosRoot:
            self.console.line()
            self.console.log(f"Exiting with exitcode={exitcode}")
            self.console.rule(
                title=":sunny:",
                style="magenta",
            )

        # Save Reports
        if self.args.text_report:
            self.console.save_text(self.args.text_report)

        if self.args.html_report:
            self.console.save_html(self.args.html_report)

        AppLogger.stop()


class YChaosRoot(YChaosSubCommand):
    """
    The YChaos root command. This class is invoked when the ychaos
    cli is invoked without any subcommands
    """

    @classmethod
    def main(cls, args: Namespace) -> Any:
        """
        Does Nothing.
        Args:
            args: Arguments

        Returns:
            0 (Zero) exitcode
        """
        return 0


# This is where it all started..
def main():
    YChaos.main(sys.argv[1:])



#  Copyright 2021, Yahoo
#  Licensed under the terms of the Apache 2.0 license. See the LICENSE file in the project root for terms

from abc import ABC, abstractmethod
from argparse import (
    ArgumentDefaultsHelpFormatter,
    ArgumentError,
    ArgumentParser,
    Namespace,
    _SubParsersAction,
)
from typing import Any, List, Optional

__all__ = ["SubCommandParsersAction", "SubCommand"]


class SubCommandParsersAction(_SubParsersAction):
    """
    Extends the `_SubParsersAction` class of Argparse that can be used for the
    action attribute of `add_subparsers`.

    ```python3
    import argparse

    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers(action=CommandSubParsersAction)
    ```

    All the parsers added to the `subparsers` using the `add_parser` method will now require a
    mandatory attribute `cls` that implements the base [SubCommand][ychaos.utils.argparse.subparsers.SubCommand] class.
    This overridden method adds a new default variable to the parser named `cls` that can be called during
    parsing arguments

    The `add_parser` method also calls the command class's `build_parser` method that takes in a parser and builds
    arguments, optionals and returns the built parser back.

    ```python3
    parser_foo = subparsers.add_parser(cls=SubCommand1)

    parser_bar = subparsers.add_parser(cls=SubCommand2)

    parsed_args = parser.parse_args()
    parsed_args.cls.main(parsed_args)
    ```

    In the above example, the two sub-commands `SubCommand1.name` and `SubCommand2.name` are
    handled by their respective classes.

    See [SubCommand][ychaos.utils.argparse.subparsers.SubCommand] for more details.
    """

    def add_parser(self, name=None, **kwargs: Any) -> ArgumentParser:
        try:
            cls = kwargs.pop("cls")
            if not issubclass(cls, SubCommand):
                raise ArgumentError(
                    self, "cls attribute must implement SubCommand class"
                )
        except KeyError as e:
            raise ArgumentError(
                self, "cls attribute is required for SubCommandParser object"
            )

        # Class command-name is given more importance
        _name = cls.name or name
        if _name is None:
            raise ArgumentError(self, "Command name cannot be None")

        # Add Aliases from Sub-Command Class
        aliases = kwargs.get("aliases", list())
        aliases.extend(cls.aliases)
        kwargs.update(dict(aliases=aliases))
        kwargs.setdefault("help", cls.help)
        kwargs.setdefault("epilog", cls.epilog)
        kwargs.setdefault("formatter_class", ArgumentDefaultsHelpFormatter)

        parser = super(SubCommandParsersAction, self).add_parser(_name, **kwargs)

        # Set Executable methods for the parser.
        parser.set_defaults(cls=cls)

        return cls.build_parser(parser)


class SubCommand(ABC):
    """
    Abstract base class for an Argparse subcommand. This is used along with the
    [SubCommandParsersAction][ychaos.utils.argparse.subparsers.SubCommandParsersAction].

    The subcommands can implement this method along with adding the action `SubCommandParsersAction` to
    `add_subparsers` method. Each class inheriting the command will implement the `build_parser` method
    to build the sub parser and return the parser.

    On parsing the arguments, the parser's `cls` attribute can be used to run the `main` function.
    """

    name: Optional[str] = None
    help: Optional[str] = None
    aliases: List[str] = list()
    epilog: Optional[str] = None

    @classmethod
    def build_parser(cls, parser: ArgumentParser) -> ArgumentParser:
        """
        Called from the SubCommandParser to add arguments, parsers to the subparser
        created from Argparse. The subclasses inheriting this class, can add arguments
        to this subparser

        Args:
            parser: Subparser object

        Returns:
            Parser object with arguments, optionals, subparsers attached.
        """
        return parser

    @classmethod
    @abstractmethod
    def main(cls, args: Namespace) -> Any:  # pragma: no cover
        """
        Defines the Program Logic to be executed when this subcommand is invoked. This method
        is called with one argument args of type Namespace which contains the parsed values.

        Args:
            args: Namespace arguments parsed from the input

        Returns:
            Any
        """
        raise NotImplementedError("execute method must be implemented by the subclass")



#  Copyright 2021, Yahoo
#  Licensed under the terms of the Apache 2.0 license. See the LICENSE file in the project root for terms
import json
import random
from types import SimpleNamespace

from ...app_logger import AppLogger
from ...testplan.attack import MachineTargetDefinition
from ...testplan.schema import TestPlan
from ...utils.dependency import DependencyUtils
from ...utils.hooks import EventHook
from .BaseExecutor import BaseExecutor

(YChaosAnsibleResultCallback,) = DependencyUtils.import_from(
    "ychaos.core.executor.common",
    ("YChaosAnsibleResultCallback",),
    raise_error=False,
    warn=False,
)

(TaskQueueManager,) = DependencyUtils.import_from(
    "ansible.executor.task_queue_manager",
    ("TaskQueueManager",),
    raise_error=False,
    warn=False,
)
(InventoryManager,) = DependencyUtils.import_from(
    "ansible.inventory.manager", ("InventoryManager",), raise_error=False, warn=False
)
(DataLoader,) = DependencyUtils.import_from(
    "ansible.parsing.dataloader", ("DataLoader",), raise_error=False, warn=False
)
(Play,) = DependencyUtils.import_from(
    "ansible.playbook.play", ("Play",), raise_error=False, warn=False
)

(VariableManager,) = DependencyUtils.import_from(
    "ansible.vars.manager", ("VariableManager",), raise_error=False, warn=False
)

(TaskResult,) = DependencyUtils.import_from(
    "ansible.executor.task_result", ("TaskResult",), raise_error=False, warn=False
)


class MachineTargetExecutor(BaseExecutor):
    """
    The executor that executes the agent on a set of Virtual machines/Bare metals
    by connecting to the hosts via SSH. The input for the executor is the testplan,
    within which, the target_type is defined as `machine`. The target_config will
    provide the list of hosts out of which random `blast_radius`% of the hosts
    is selected for attack.

    The following are the valid hooks to this executor

    ## Valid Hooks

    === "on_start"
        Called when starting to run the Ansible playbook.

        ```python
            def callable_hook(): ...
        ```

    === "on_no_targets_found"
        Called when no targets are found to attack. Either blast radius is
        too low or there are no hosts in the attack list.

        ```python
            def callable_hook(): ...
        ```

    === "on_target_unreachable"
        Called when a target becomes unreachable during the Ansible play

        ```python
            def callable_hook(result: AnsibleResult): ...
        ```

    === "on_target_passed"
        Called when an ansible task is passed

        ```python
            def callable_hook(result: AnsibleResult): ...
        ```

    === "on_target_failed"
        Called when an ansible task is failed

        ```python
            def callable_hook(result: AnsibleResult): ...
        ```

    === "on_end"
        Called after the end of Ansible playbook run

        ```python
            def callable_hook(result): ...
        ```

    === "on_error"
        Called when an exception is raised

        ```python
            def callable_hook(e: Exception): ...
        ```
    """

    __target_type__ = "machine"

    # __taskresult_callable__ = EventHook.Callable(TaskResult)

    __hook_events__ = {
        "on_no_targets_found": EventHook.CallableType(),
        "on_start": EventHook.CallableType(),
        "on_target_unreachable": EventHook.CallableType(TaskResult),
        "on_target_failed": EventHook.CallableType(TaskResult),
        "on_target_passed": EventHook.CallableType(TaskResult),
        "on_error": EventHook.CallableType(Exception),
        "on_end": EventHook.CallableType(TaskResult),
    }

    def __init__(self, testplan: TestPlan, *args, **kwargs):
        super(MachineTargetExecutor, self).__init__(testplan)

        # Selects a `blast_radius`% of hosts at random from the
        # effective hosts and uses it as the target hosts for the attack
        self._compute_target_hosts()

        self.ansible_context = SimpleNamespace()

        self.logger = AppLogger.get_logger(self.__class__.__name__)

    def _compute_target_hosts(self):
        target_defn: MachineTargetDefinition = self.testplan.attack.get_target_config()
        effective_hosts = target_defn.get_effective_hosts()
        self.target_hosts = random.sample(
            effective_hosts, target_defn.blast_radius * len(effective_hosts) // 100
        )

    def prepare(self):
        self.ansible_context.loader = DataLoader()
        self.ansible_context.results_callback = YChaosAnsibleResultCallback(
            hooks={
                k: v
                for k, v in self.hooks.items()
                if k in YChaosAnsibleResultCallback.__hook_events__
            }
        )

        # Hosts to be in comma separated string
        hosts = ",".join(self.testplan.attack.get_target_config().get_effective_hosts())
        if len(self.testplan.attack.get_target_config().get_effective_hosts()) == 1:
            hosts += ","

        self.ansible_context.inventory = InventoryManager(
            loader=self.ansible_context.loader, sources=hosts
        )
        self.ansible_context.variable_manager = VariableManager(
            loader=self.ansible_context.loader, inventory=self.ansible_context.inventory
        )

        self.ansible_context.tqm = TaskQueueManager(
            inventory=self.ansible_context.inventory,
            variable_manager=self.ansible_context.variable_manager,
            loader=self.ansible_context.loader,
            passwords=dict(vault_pass=None),
            stdout_callback=self.ansible_context.results_callback,
        )

        self.ansible_context.play_source = dict(
            name="YChaos Ansible Play",
            hosts=",".join(self.target_hosts),
            remote_user=self.testplan.attack.get_target_config().ssh_config.user,
            connection="ssh",
            strategy="free",
            gather_facts="no",
            ignore_unreachable="no",
            tasks=[
                dict(
                    name="Check current working directory",
                    action=dict(module="command", args=dict(cmd="pwd")),
                    register="result_pwd",
                    changed_when="false",
                ),
                dict(
                    name="Check if python3 installed",
                    action=dict(module="command", args=dict(cmd="which python3")),
                    register="result_which_python3",
                    changed_when="false",
                    failed_when=[
                        # Fail if python3 not installed on the target
                        "result_which_python3.rc != 0"
                    ],
                ),
                dict(
                    name="Create a virtual environment",
                    register="result_pip",
                    action=dict(
                        module="pip",
                        chdir="{{result_pwd.stdout}}",
                        name="pip",
                        state="latest",
                        virtualenv="ychaos_env",
                        virtualenv_command="{{result_which_python3.stdout}} -m venv",
                    ),
                    failed_when=[
                        # Failed in these following reasons
                        # 1. pip not installed
                        # 2. Failed to installed latest pip version
                        "result_pip.state == 'absent'"
                    ],
                    vars=dict(
                        ansible_python_interpreter="{{result_which_python3.stdout}}"
                    ),
                ),
                dict(
                    name="Install ychaos[agents]",
                    register="result_pip_install_ychaos_agents",
                    action=dict(
                        module="pip",
                        chdir="{{result_pwd.stdout}}",
                        name="ychaos[agents]",
                        virtualenv="ychaos_env",
                    ),
                    failed_when=[
                        # Failed if unable to install ychaos[agents]
                        "result_pip_install_ychaos_agents.state == 'absent'"
                    ],
                    vars=dict(
                        ansible_python_interpreter="{{result_which_python3.stdout}}"
                    ),
                ),
                dict(
                    name="Create a workspace directory for storing local report files",
                    register="result_create_workspace",
                    action=dict(
                        module="file",
                        path="{{result_pwd.stdout}}/ychaos_ws",
                        state="directory",
                        mode="0755",
                    ),
                ),
                dict(
                    name="Copy testplan from local to remote",
                    register="result_testplan_file",
                    action=dict(
                        module="copy",
                        content=json.dumps(
                            self.testplan.to_serialized_dict(), indent=4
                        ),
                        dest="{{result_create_workspace.path}}/testplan.json",
                    ),
                ),
                dict(
                    name="Run YChaos Agent",
                    ignore_errors="yes",
                    action=dict(
                        module="shell",
                        cmd=" ".join(
                            [
                                "source {{result_pip.virtualenv}}/bin/activate",
                                "&&",
                                "ychaos --log-file {{result_create_workspace.path}}/ychaos.log",
                                "agent attack --testplan {{result_testplan_file.dest}} --attack-report-yaml {{result_create_workspace.path}}/attack_report.yaml",
                            ]
                        ),
                    ),
                ),
                dict(
                    name="Zip workspace directory",
                    action=dict(
                        module="community.general.archive",
                        path="{{result_create_workspace.path}}",
                        dest="{{result_create_workspace.path}}/ychaos.zip",
                        format="zip",
                    ),
                ),
                dict(
                    name="Copy Workspace directory to local",
                    action=dict(
                        module="fetch",
                        src="{{result_create_workspace.path}}/ychaos.zip",
                        dest=str(
                            self.testplan.attack.get_target_config().report_dir.resolve()
                        )
                        + "/ychaos_{{inventory_hostname}}.zip",
                    ),
                ),
                dict(
                    name="Delete YChaos Workspace on host",
                    action=dict(
                        module="file",
                        path="{{result_create_workspace.path}}",
                        state="absent",
                    ),
                ),
                dict(
                    name="Delete Virtual environment",
                    action=dict(
                        module="file",
                        path="{{result_pip.virtualenv}}",
                        state="absent",
                    ),
                ),
            ],
        )

    def execute(self) -> None:
        self.prepare()

        if len(self.target_hosts) == 0:
            self.execute_hooks("on_no_targets_found")
            return

        play = Play().load(
            self.ansible_context.play_source,
            variable_manager=self.ansible_context.variable_manager,
            loader=self.ansible_context.loader,
        )

        # Create Report Directory
        import os

        os.makedirs(
            self.testplan.attack.get_target_config().report_dir.resolve(), exist_ok=True
        )

        try:
            self.execute_hooks("on_start")
            result = self.ansible_context.tqm.run(play)
            self.execute_hooks("on_end", result)
        except Exception as e:
            self.execute_hooks("on_error", e)
        finally:
            self.ansible_context.tqm.cleanup()
            if self.ansible_context.loader:  # pragma: no cover
                self.ansible_context.loader.cleanup_all_tmp_files()



#  Copyright 2021, Yahoo
#  Licensed under the terms of the Apache 2.0 license. See the LICENSE file in the project root for terms
import math
import warnings
from datetime import datetime, timedelta
from multiprocessing import Pool, cpu_count
from queue import LifoQueue

from pydantic import Field, validate_arguments

from ...utils.builtins import BuiltinUtils
from ...utils.dependency import DependencyUtils
from ..agent import Agent, AgentMonitoringDataPoint, TimedAgentConfig
from ..utils.annotations import log_agent_lifecycle

__all__ = ["CPUBurnConfig", "CPUBurn"]


def _burn(end: datetime) -> None:
    """
    A simple timed infinite loop that is supposed to run on all the processors
    The while loop should contain a CPU intensive operation that can burn CPU during its run.

    Args:
        end: when to stop the execution of the agent

    Returns:
        None
    """
    while True:
        if datetime.now() >= end:
            _ = (
                22112 ** 47373
            )  # Random Number Powered by Random Number (Make sure to test this completes execution in a second)
            break


class CPUBurnConfig(TimedAgentConfig):
    """
    Defines the CPU Burn configuration to initiate a CPU burn attack. The framework will
    attack only a percentage of cores that is defined in the `cores_pct` attribute. By default,
    `cores_pct` is set to 100, implying all the CPU cores are targeted simultaneously.
    """

    name = "cpu_burn"
    description = "This agent is responsible to consume the CPU resources for the `duration` amount of seconds."

    cores_pct: float = Field(
        default=100,
        description=(
            "Percentage of all the cores to use. "
            "This will burn CPU only on a percentage of cores available in the system"
        ),
        ge=0,
        le=100,
    )

    def effective_cpu_count(self) -> int:
        """
        Calculates the number of cores to be used from the cores_pct information
        Returns:
            number of cores that fits in the `cores_pct` percentage
        """
        return math.floor(self.cores_pct * cpu_count() / 100)


class CPUBurn(Agent):
    _psutil = None

    @validate_arguments
    def __init__(self, config: CPUBurnConfig):
        super(CPUBurn, self).__init__(config)

        self._psutil = DependencyUtils.import_module("psutil", raise_error=False)

        if self._psutil is None:
            warnings.warn(
                "psutil is not installed. The agent cannot monitor the metrics related to the system."
                "You can install psutil if you are interested in getting the system data",
                category=ImportWarning,
            )

    def monitor(self) -> LifoQueue:
        # If `psutil` is installed, the agent will be able to monitor the system metrics within
        # the agent. If the `psutil` package is not installed, the agent will not able to monitor
        # the system metrics. The agent will not throw an error because of a missing package.
        # Instead the output of the monitoring data will be NaN.

        cpu_usage = BuiltinUtils.Float.NAN
        if self._psutil is not None and hasattr(self._psutil, "cpu_percent"):
            cpu_usage = sum(self._psutil.cpu_percent(0.5, True)) // cpu_count()  # type: ignore
            # TODO: Reason for type ignore - https://github.com/python/mypy/issues/1424

        self._status.put(
            AgentMonitoringDataPoint(
                data=dict(
                    cpu_count=self.config.effective_cpu_count(), cpu_usage=cpu_usage
                ),
                state=self.current_state,
            )
        )
        return self._status

    @log_agent_lifecycle
    def setup(self) -> None:
        super(CPUBurn, self).setup()

    @log_agent_lifecycle
    def run(self) -> None:
        super(CPUBurn, self).run()
        end = datetime.now() + timedelta(seconds=self.config.duration)

        if self.config.effective_cpu_count() == 0:
            return

        process_pool = Pool(self.config.effective_cpu_count())
        process_pool.map_async(_burn, (end,) * self.config.effective_cpu_count())

    @log_agent_lifecycle
    def teardown(self) -> None:
        super(CPUBurn, self).teardown()



#  Copyright 2021, Yahoo
#  Licensed under the terms of the Apache 2.0 license. See the LICENSE file in the project root for terms
import functools
import time
from queue import LifoQueue

from ..agent import Agent, AgentConfig, TimedAgentConfig
from ..utils.annotations import log_agent_lifecycle

NoOpAgentConfig = functools.partial(AgentConfig, name="no_op")


class NoOpAgent(Agent):
    """
    An agent that does nothing
    """

    def __init__(self, config):
        assert isinstance(config, AgentConfig)
        super(NoOpAgent, self).__init__(config)

    def monitor(self) -> LifoQueue:  # pragma: no cover
        super(NoOpAgent, self).monitor()
        return self._status

    @log_agent_lifecycle
    def setup(self) -> None:
        super(NoOpAgent, self).setup()

    @log_agent_lifecycle
    def run(self) -> None:
        super(NoOpAgent, self).run()

    @log_agent_lifecycle
    def teardown(self) -> None:
        super(NoOpAgent, self).teardown()


NoOpTimedAgentConfig = functools.partial(TimedAgentConfig, name="no_op_timed")


class NoOpTimedAgent(NoOpAgent):  # pragma: no cover
    """
    A time constrained agent that does nothing for `duration` seconds.
    """

    def __init__(self, config):
        assert isinstance(config, TimedAgentConfig)
        super(NoOpTimedAgent, self).__init__(config)

    @log_agent_lifecycle
    def run(self) -> None:
        super(NoOpTimedAgent, self).run()
        time.sleep(self.config.duration)



#  Copyright 2021, Yahoo
#  Licensed under the terms of the Apache 2.0 license. See the LICENSE file in the project root for terms
import time
from datetime import datetime, timedelta
from typing import Any, Dict

import requests
from pydantic import validate_arguments
from requests import HTTPError, Session, Timeout

from ....testplan.verification import SDv4Verification
from ..data import VerificationData, VerificationStateData
from .BaseVerificationPlugin import BaseVerificationPlugin


class SDv4VerificationPlugin(BaseVerificationPlugin):

    __verification_type__ = "sdv4"

    @validate_arguments
    def __init__(
        self,
        config: SDv4Verification,
        state_data: VerificationData = VerificationData.parse_obj(dict()),
    ):
        super(SDv4VerificationPlugin, self).__init__(config, state_data)
        self._session = Session()

    def _get_bearer_token(self):
        response = requests.get(
            url=f"{self.config.sd_api_url}/v4/auth/token",
            params=dict(api_token=self.config.sd_api_token.get_secret_value()),
            timeout=10,
        )
        response.raise_for_status()
        return response.json().get("token")

    def _monitor_job(self, sdv4_event: Dict[Any, Any]):

        _job_monitor_end_time = datetime.now() + timedelta(
            seconds=self.config.job_timeout
        )

        event_id = sdv4_event.get("id")
        self.logger.info(
            pipeline_id=self.config.pipeline_id,
            job_name=self.config.job_name,
            event_id=event_id,
            msg="Monitoring SDv4 job",
        )
        while datetime.now() < _job_monitor_end_time:
            sdv4_event_get_response = self._session.get(
                url=f"{self.config.sd_api_url}/v4/events/{event_id}/builds"
            )
            sdv4_event_get_response.raise_for_status()

            sdv4_event_build = sdv4_event_get_response.json()
            build = sdv4_event_build[0]
            if build["status"] in (
                "ABORTED",
                "FAILURE",
                "BLOCKED",
                "UNSTABLE",
                "FROZEN",
            ):
                self.logger.info(
                    status=build["status"], msg="SDv4 job verification failed"
                )
                return VerificationStateData(
                    rc=2,
                    type=self.__verification_type__,
                    data=dict(
                        event_id=build["eventId"],
                        status=build["status"],
                        status_message=build.get("statusMessage", None),
                        job_id=build["jobId"],
                    ),
                )
            elif build["status"] in ("SUCCESS",):
                self.logger.info(
                    status=build["status"], msg="SDv4 job verification successful"
                )
                return VerificationStateData(
                    rc=0,
                    type=self.__verification_type__,
                    data=dict(
                        event_id=build["eventId"],
                        status=build["status"],
                        status_message=build.get("statusMessage", None),
                        job_id=build["jobId"],
                    ),
                )
            else:
                # Status = ("CREATED", "QUEUED", "RUNNING")
                # Wait for 60 seconds and check for the status of the build again
                time.sleep(60)

    def run_verification(self) -> VerificationStateData:
        self.logger.info(msg="Starting SDv4 job verification")
        try:
            bearer_token = self._get_bearer_token()

            self._session.headers["Authorization"] = bearer_token
            self._session.headers["accept"] = "application/json"
            self._session.headers["Content-Type"] = "application/json"

            self.logger.info(
                pipeline_id=self.config.pipeline_id,
                job_name=self.config.job_name,
                msg="Starting SDv4 job",
            )
            sdv4_event_post_response = self._session.post(
                url=f"{self.config.sd_api_url}/v4/events",
                json=dict(
                    causeMessage="YChaos SDv4 Verification",
                    creator=dict(name="ychaos", username="ychaos"),
                    pipelineId=self.config.pipeline_id,
                    startFrom=self.config.job_name,
                ),
            )
            sdv4_event_post_response.raise_for_status()

            self.logger.info(
                pipeline_id=self.config.pipeline_id,
                job_name=self.config.job_name,
                msg="Started SDv4 job",
            )
            time.sleep(2)  # 2 second sleep for SDv4 to start the Job
            return self._monitor_job(sdv4_event_post_response.json())

        except HTTPError as http_error:
            self.logger.info(msg="SDv4 job verification failed")
            return VerificationStateData(
                rc=1,
                type=self.__verification_type__,
                data=dict(
                    url=http_error.response.url,
                    status_code=http_error.response.status_code,
                    json=http_error.response.json(),
                    error=http_error.__class__.__name__,
                ),
            )
        except Timeout as timeout_error:
            self.logger.info(msg="SDv4 job verification failed")
            return VerificationStateData(
                rc=1,
                type=self.__verification_type__,
                data=dict(error=timeout_error.__class__.__name__),
            )



#  Copyright 2021, Yahoo
#  Licensed under the terms of the Apache 2.0 license. See the LICENSE file in the project root for terms
from abc import ABC
from argparse import ArgumentParser, Namespace
from pathlib import Path
from typing import Any, Optional

from pydantic import ValidationError
from rich.console import Console
from rich.panel import Panel

from ..testplan.schema import TestPlan
from ..utils.argparse import SubCommand
from .exceptions import YChaosCLIError


class YChaosArgumentParser(ArgumentParser):
    """
    YChaosArgumentParser serves as a special parser to parse the
    CLI commands for ychaos. It provides utility methods to parse arguments,
    pass subcommand implementation and run a subcommand. It also handles
    the CLI exception gracefully by running the `handle` method of the exception raised.
    """

    def __init__(self, *args, **kwargs):
        """
        Initialize a YChaos Argument parser.

        Args:
            **kwargs: `__root__` for the root execution when no subcommand is thrown.
        """
        # __root__ SubCommand class determines the root of the YChaos
        # i.e. when YChaos CLI is called without any arguments
        self.__root__ = kwargs.pop("__root__", None)

        super(YChaosArgumentParser, self).__init__(*args, **kwargs)

    def parse_args(self, args=None, namespace=None) -> Namespace:  # type: ignore
        args = super(YChaosArgumentParser, self).parse_args(args, namespace)
        if not hasattr(args, "cls"):
            assert issubclass(self.__root__, SubCommand)
            args.cls = self.__root__

        return args

    def run_command(self, args: Namespace) -> int:
        """
        Run a subcommand from YChaos CLI
        """
        try:
            exitcode = args.cls.main(args)
        except YChaosCLIError as e:
            e.handle()
            exitcode = e.exitcode

        return exitcode


class YChaosSubCommand(SubCommand, ABC):

    _exitcode = 0

    def __init__(self, **kwargs):
        assert kwargs.pop("cls") == self.__class__

        self.app = kwargs.pop("app")
        self.console: Console = self.app.console

    def set_exitcode(self, exitcode=0):
        self._exitcode = exitcode

    @classmethod
    def main(cls, args: Namespace) -> Any:
        args.app.console.log("This command does nothing..")
        return 0


class YChaosTestplanInputSubCommand(YChaosSubCommand, ABC):
    @classmethod
    def build_parser(cls, parser: ArgumentParser) -> ArgumentParser:
        parser.add_argument(
            "-t",
            "--testplan",
            type=Path,
            required=True,
            help="The testplan path. This can be relative path from where the CLI is initiated",
            metavar="path",
        )

        return parser

    def get_validated_test_plan(self, path: Path) -> Optional[TestPlan]:
        self.console.log("Getting Test plan")
        self.console.line()
        try:
            testplan = TestPlan.load_file(path=path)
            return testplan
        except ValidationError as validation_error:
            self.set_exitcode(1)
            self.console.print(
                Panel.fit(str(validation_error), title="Validation Error", style="red")
            )
        except IsADirectoryError as is_directory:
            self.set_exitcode(1)
            self.console.print(
                ":file_folder: The input path ({path}) is not a valid testplan file".format(
                    path=path
                )
            )
        except FileNotFoundError as file_not_found_error:
            self.set_exitcode(1)
            self.console.print(
                ":mag: {file} [italic]not found[/italic]".format(file=str(path)),
                style="indian_red",
            )
        return None


class YChaosCLIHook(ABC):
    """
    The Base CLI Hook. Implement the `__call__` method with appropriate signature
    for each of the hook type. The `__call__` method gets called from the EventHook
    methods whenever an event is occurred.
    """

    def __init__(self, app, *args, **kwargs):
        self.app = app
        self.console = app.console

        # Determines if this hooks is active and callable.
        self.active = True



#  Copyright 2021, Yahoo
#  Licensed under the terms of the Apache 2.0 license. See the LICENSE file in the project root for terms
import json
from datetime import datetime
from typing import Any, Dict, List, Optional, Union

from pydantic import BaseModel, Field

from ...testplan import SystemState
from ...testplan.verification import VerificationType


class VerificationStateData(BaseModel):
    rc: int = Field(
        ...,
        description="The return code of the verification plugin. 0, +x and -x respectively indicate success, failure and error.",
    )
    timestamp: datetime = Field(
        default_factory=datetime.utcnow,
        description="The timestamp at which this state data was recorded.",
    )
    type: VerificationType = Field(
        ...,
        description="The type of verification plugin that ran to record this state data",
    )

    data: Union[Dict[Any, Any], List[Any]] = Field(
        default=dict(),
        description="Plugin level data that can be consumed by the respective plugin",
    )

    class Config:
        json_encoders = {datetime: lambda v: int(v.timestamp())}


class VerificationData(BaseModel):
    __root__: Dict[SystemState, Optional[VerificationStateData]] = Field(
        default=dict.fromkeys(SystemState, None)
    )

    def encoded_dict(self):
        return {
            state.value: json.loads(state_data.json()) if state_data else None
            for state, state_data in self.__root__.items()
        }

    @classmethod
    def parse_obj(
        cls, obj: Dict[SystemState, Optional[VerificationStateData]]
    ) -> "VerificationData":
        for system_state in SystemState:
            obj.setdefault(system_state, None)
        return super(VerificationData, cls).parse_obj(obj)

    def add_data(
        self,
        system_state: SystemState,
        state_data: Optional[VerificationStateData],
        overwrite=False,
    ):
        if overwrite:
            self.__root__[system_state] = state_data
        else:
            self.__root__.setdefault(system_state, state_data)

    def replace_data(
        self, system_state: SystemState, state_data: VerificationStateData
    ):
        self.add_data(system_state, state_data, overwrite=True)

    def is_data_present(self, system_state: SystemState):
        return self.get_data(system_state) is not None

    def get_data(self, system_state: SystemState):
        return self.__root__[system_state]



#  Copyright 2021, Yahoo
#  Licensed under the terms of the Apache 2.0 license. See the LICENSE file in the project root for terms

from argparse import ArgumentParser, Namespace
from pathlib import Path
from typing import Any, List

from pydantic import ValidationError
from rich.panel import Panel

from ...testplan.validator import TestPlanValidator
from .. import YChaosSubCommand

__all__ = ["TestPlanValidatorCommand"]


class TestPlanValidatorCommand(YChaosSubCommand):
    """
    Test plan validator subcommand is used to validate whether the testplan
    files adheres to the YChaos Test plan schema.
    """

    name = "validate"
    help = "Validate YChaos Test plans"

    @classmethod
    def build_parser(cls, parser: ArgumentParser) -> ArgumentParser:
        parser.add_argument(
            "paths",
            type=Path,
            nargs="+",
            help="Space separated list of file/directory paths to validate",
            metavar="path",
        )

        return parser

    def __init__(self, **kwargs):
        super(TestPlanValidatorCommand, self).__init__(**kwargs)

        self.paths: List[Path] = kwargs.pop("paths")

    def resolve_validation_paths(self) -> List[Path]:
        file_types = ("json", "yaml", "yml")
        resolved_filepaths: List[Path] = list()
        for input_path in self.paths:
            if input_path.resolve().is_dir():
                for file_type in file_types:
                    resolved_filepaths.extend(
                        list(input_path.glob(f"**/*.{file_type}"))
                    )
            else:
                resolved_filepaths.append(input_path)

        return resolved_filepaths

    def do_testplans_validation(self):
        self.console.log("Getting Test plans")

        resolved_filepaths = self.resolve_validation_paths()

        if len(resolved_filepaths) == 0:
            self.console.line()
            self.console.print(
                ":open_file_folder: No Test plans found", style="orange3"
            )
            return

        self.console.log("Validating Test plans")
        self.console.line()

        for file in sorted(resolved_filepaths):
            try:
                TestPlanValidator.validate_file(file)
                self.console.print(
                    f":white_check_mark: {file}",
                    style="green",
                )

            except ValidationError as validation_error:
                self.set_exitcode(1)

                self.console.print("")
                self.console.print(f":exclamation: {file}", style="bold red")
                self.console.print(
                    Panel.fit(
                        str(validation_error), title="Validation Error", style="red"
                    )
                )
                self.console.print("")

            except FileNotFoundError as file_not_found_error:
                self.set_exitcode(1)
                self.console.print(
                    f":mag: {file} [italic]not found[/italic]",
                    style="indian_red",
                )

    @classmethod
    def main(cls, args: Namespace) -> Any:
        validator = cls(**vars(args))
        validator.do_testplans_validation()

        return validator._exitcode



#  Copyright 2021, Yahoo
#  Licensed under the terms of the Apache 2.0 license. See the LICENSE file in the project root for terms
from datetime import datetime
from typing import Any, Dict, List, Optional

from pydantic import BaseModel, validate_arguments

from ....testplan.verification import OpenTSDBVerification
from ..data import VerificationData, VerificationStateData
from .BaseVerificationPlugin import RequestVerificationPlugin


class OpenTSDBVerificationPlugin(RequestVerificationPlugin):
    class OpenTSDBVerificationData(BaseModel):
        url: str  # Making this string since this is built with only trusted data
        status_code: Optional[int]
        data: Dict[datetime, float]

    __verification_type__ = "tsdb"

    @validate_arguments
    def __init__(
        self,
        config: OpenTSDBVerification,
        state_data: VerificationData = VerificationData.parse_obj(dict()),
    ):
        super(OpenTSDBVerificationPlugin, self).__init__(config, state_data)
        self._session = self._build_session()

    def run_verification(self) -> VerificationStateData:
        _rc = 0

        tsdb_response = self._session.request(
            self.config.method,
            url=str(self.config.url),
            timeout=self.config.timeout / 1000,
        )

        if tsdb_response.status_code != 200:
            return VerificationStateData(
                rc=-1,
                type=self.__verification_type__,
                data=self.OpenTSDBVerificationData(
                    url=self.config.url,
                    status_code=tsdb_response.status_code,
                    data=dict(),
                ).dict(),
            )

        else:
            if len(self.config.criteria) != 0:
                _rc = self.validate_criteria(tsdb_response.json())
            else:
                # TODO: Verify State Bound Critieria
                pass

        return VerificationStateData(
            rc=_rc, type=self.__verification_type__, data=tsdb_response.json()
        )

    def validate_criteria(self, response_data: List[Dict[str, Any]]):
        for _query_data in response_data:

            for _criteria in self.config.criteria:

                _aggregated_data = _criteria.aggeragator.metadata.aggregate(  # type: ignore
                    _query_data["dps"]
                )

                for _condition in _criteria.conditionals:
                    if _condition.comparator.metadata.compare(  # type: ignore
                        _aggregated_data, _condition.value
                    ):
                        break
                else:
                    return 1

        else:
            return 0



#  Copyright 2021, Yahoo
#  Licensed under the terms of the Apache 2.0 license. See the LICENSE file in the project root for terms

import logging
from logging import DEBUG, ERROR, INFO, WARNING, Logger
from typing import Optional, Set


class StructLogger(Logger):
    def __init__(self, name):
        super().__init__(name)
        logging.setLoggerClass(self.__class__)
        self._binder = dict()

    def _build_msg(self, msg="", **kwargs) -> str:
        bind_msg = " ".join([f"{k}={v}" for k, v in self._binder.items()])
        kwargs_msg = " ".join([f"{k}={v}" for k, v in kwargs.items()])

        return f"{bind_msg} {kwargs_msg} {msg}"

    def getChild(self, suffix: str, bind_parent_attributes=False) -> "StructLogger":
        """
        Get Child Logger and bind Parent attributes to child if needed.
        Args:
            suffix: Suffix to be added to the child logger
            bind_parent_attributes: Bind Parent Attributes to Child if needed
        Returns:
            Child Struct Logger
        """
        _logger: StructLogger = super(StructLogger, self).getChild(suffix)  # type: ignore
        if bind_parent_attributes:
            assert isinstance(_logger, StructLogger)
            _logger.bind(**self._binder)

        return _logger

    def debug(self, msg="", *args, **kwargs) -> None:
        level = DEBUG
        exc_info, extra, stack_info = (
            kwargs.pop("exc_info", None),
            kwargs.pop("extra", None),
            kwargs.pop("stack_info", False),
        )
        if self.isEnabledFor(level):
            self._log(
                level,
                self._build_msg(msg, **kwargs),
                args,
                exc_info=exc_info,
                extra=extra,
                stack_info=stack_info,
            )

    def info(self, msg="", *args, **kwargs) -> None:
        level = INFO
        exc_info, extra, stack_info = (
            kwargs.pop("exc_info", None),
            kwargs.pop("extra", None),
            kwargs.pop("stack_info", False),
        )
        if self.isEnabledFor(level):
            self._log(
                level,
                self._build_msg(msg, **kwargs),
                args,
                exc_info=exc_info,
                extra=extra,
                stack_info=stack_info,
            )

    def error(self, msg="", *args, **kwargs) -> None:
        level = ERROR
        exc_info, extra, stack_info = (
            kwargs.pop("exc_info", None),
            kwargs.pop("extra", None),
            kwargs.pop("stack_info", False),
        )
        if self.isEnabledFor(level):
            self._log(
                level,
                self._build_msg(msg, **kwargs),
                args,
                exc_info=exc_info,
                extra=extra,
                stack_info=stack_info,
            )

    def warning(self, msg="", *args, **kwargs) -> None:
        level = WARNING
        exc_info, extra, stack_info = (
            kwargs.pop("exc_info", None),
            kwargs.pop("extra", None),
            kwargs.pop("stack_info", False),
        )
        if self.isEnabledFor(level):
            self._log(
                level,
                self._build_msg(msg, **kwargs),
                args,
                exc_info=exc_info,
                extra=extra,
                stack_info=stack_info,
            )

    def exception(self, msg="", *args, **kwargs) -> None:
        level = ERROR
        exc_info, extra, stack_info = (
            kwargs.pop("exc_info", True),
            kwargs.pop("extra", None),
            kwargs.pop("stack_info", False),
        )
        if self.isEnabledFor(level):
            self._log(
                level,
                self._build_msg(msg, **kwargs),
                args,
                exc_info=exc_info,
                extra=extra,
                stack_info=stack_info,
            )

    def bind(self, **kwargs) -> None:
        self._binder = kwargs

    def unbind(self, unbind_keys: Optional[Set] = None) -> None:
        """
        unbind certain/all keys

        Args:
            unbind_keys: Set of keys to unbind Or None to unbind all keys
        """
        if unbind_keys:
            for k in unbind_keys:
                self._binder.pop(k, None)
        else:
            self._binder.clear()



#  Copyright 2021, Yahoo
#  Licensed under the terms of the Apache 2.0 license. See the LICENSE file in the project root for terms

import os
from datetime import datetime, timedelta, timezone
from logging import Logger
from queue import Queue
from threading import Thread
from time import sleep
from typing import Dict, List, Optional

from pydantic import BaseModel

from ..app_logger import AppLogger
from ..testplan.attack import AttackMode
from ..testplan.schema import TestPlan
from ..utils.hooks import EventHook
from .agent import Agent, AgentState


class ConfiguredAgent:
    """
    Class to hold the configured Agent
    """

    def __init__(
        self, agent: Agent, start_time: Optional[datetime], end_time: Optional[datetime]
    ):
        """
        Initialize ConfiguredAgent
        Args:
            agent: Agent object
            start_time: Agent execution start time
            end_time: Agent execution end time
        """
        self.agent: Agent = agent
        self.start_time = start_time
        self.end_time = end_time
        self.agent_start_thread: Optional[Thread] = None
        self.agent_teardown_thread: Optional[Thread] = None


class Coordinator(EventHook):
    """
    The coordinator is responsible for setting up the chaos agents,
    running the agents and monitor the agent currently being run. It also
    takes care of completing the attack by bringing back to the system to its original
    state.

    ## Coordinator Event Hooks

    === "on_attack_start"
        called when attack is started on the host
        ```python
            def callable_hook(): ...
        ```

    === "on_attack_completed"
        called when attack is completed
        ```python
            def callable_hook(): ...
        ```

    === "on_each_agent_start"
        called when a Agent start executing
        ```python
            def callable_hook(agent_name: str): ...
        ```

    === "on_each_agent_teardown"
        called when a Agent Teardown is started
        ```python
            def callable_hook(agent_name: str): ...
        ```

    === "on_each_agent_stop"
        called when a Agent completes its execution and teardown step
        ```python
            def callable_hook(agent_name: str): ...
        ```
    ---
    """

    __hook_events__ = {
        "on_attack_start": EventHook.CallableType(),
        "on_attack_completed": EventHook.CallableType(),
        "on_each_agent_start": EventHook.CallableType(str),
        "on_each_agent_teardown": EventHook.CallableType(str),
        "on_each_agent_stop": EventHook.CallableType(str),
    }
    DEFAULT_DURATION = 3
    THREAD_TIMEOUT = 300

    def __init__(self, test_plan: TestPlan):
        super(Coordinator, self).__init__()
        self.test_plan: TestPlan = test_plan
        self.configured_agents: List[ConfiguredAgent] = []
        self.attack_end_time: Optional[datetime] = None
        self.attack_start_time: Optional[datetime] = None
        self.exit_code = 0
        self.log: Logger = AppLogger.get_logger(__name__)

    def configure_agent_in_test_plan(self) -> List[ConfiguredAgent]:
        """
        Configure all the Agents as specified in the test plan
        Returns:
            list of configured agents
        """

        for agent in self.test_plan.attack.agents:
            next_start_time = datetime.now(timezone.utc)
            configured_agent: Agent

            if self.test_plan.attack.mode.value == AttackMode.SEQUENTIAL.value and len(
                self.configured_agents
            ):
                assert self.configured_agents[-1].end_time is not None
                next_start_time = self.configured_agents[-1].end_time

            agent_config = agent.type.metadata.schema(**agent.config)  # type: ignore
            configured_agent = agent.type.metadata.agent_defn(agent_config)  # type: ignore
            start_time: datetime = next_start_time + timedelta(
                seconds=configured_agent.config.start_delay
            )
            end_time: datetime = start_time + timedelta(
                seconds=getattr(
                    configured_agent.config, "duration", self.DEFAULT_DURATION
                )
            )

            self.configured_agents.append(
                ConfiguredAgent(
                    configured_agent, start_time=start_time, end_time=end_time
                )
            )
        if self.test_plan.attack.mode.value != AttackMode.SEQUENTIAL.value:
            self.configured_agents.sort(
                key=lambda current_agent: current_agent.end_time  # type: ignore
            )
            self.attack_end_time = self.configured_agents[-1].end_time
            self.configured_agents.sort(
                key=lambda current_agent: current_agent.start_time  # type: ignore
            )
            self.attack_start_time = self.configured_agents[0].start_time
        else:
            self.attack_start_time = self.configured_agents[0].start_time
            self.attack_end_time = self.configured_agents[-1].end_time
        return self.configured_agents

    def get_exit_status(self) -> int:
        return self.exit_code

    def get_next_agent_for_attack(self) -> Optional[ConfiguredAgent]:
        """
        Get the next Agent for execution as configured in test plan
        Returns:
            An Agent or None
        """
        current_time: datetime = datetime.now(timezone.utc)
        for configured_agent in self.configured_agents:
            assert configured_agent.start_time is not None
            assert configured_agent.end_time is not None
            if (
                configured_agent.agent.current_state == AgentState.INIT
                and current_time > configured_agent.start_time
            ):
                try:
                    configured_agent.agent.setup()
                except Exception as e:
                    configured_agent.agent.exception.put(e)
                    configured_agent.agent.advance_state(AgentState.ERROR)
                    self.exit_code = 1
                    break
                else:
                    return configured_agent
        return None

    def get_next_agent_for_teardown(self) -> Optional[ConfiguredAgent]:
        """
        Get the next Agent for teardown as configured in test plan
        Returns:
            An Agent or None
        """
        current_time: datetime = datetime.now(timezone.utc)
        for configured_agent in self.configured_agents:
            assert configured_agent.start_time is not None
            assert configured_agent.end_time is not None
            if (
                configured_agent.agent.current_state == AgentState.RUNNING
                and current_time > configured_agent.end_time
                and not configured_agent.agent_teardown_thread
            ):
                return configured_agent
        return None

    def check_for_failed_agents(self, agent: Optional[Agent] = None) -> bool:
        """
        check if any Agent has error
        Args:
            agent: check only this Agent has error

        Returns:
            True if Agent has error else False
        """
        for configured_agent in self.configured_agents:
            if agent and agent != configured_agent.agent:
                continue
            if (
                configured_agent.agent.current_state == AgentState.ERROR
                or not configured_agent.agent.exception.empty()
            ):
                configured_agent.agent.advance_state(AgentState.ERROR)
                configured_agent.agent.preserved_state.has_error = True
                return True
        return False

    def stop_all_running_agents_in_sync(self):
        """
        waits for all agents to complete teardown step
        """
        for configured_agent in self.configured_agents:
            if (
                configured_agent.agent.current_state == AgentState.SETUP
                or configured_agent.agent.current_state == AgentState.INIT
            ):
                configured_agent.agent.advance_state(AgentState.SKIPPED)
            elif configured_agent.agent.current_state == AgentState.ERROR:
                configured_agent.agent.preserved_state.has_error = True
                self.exit_code = 1
            elif (
                configured_agent.agent_start_thread
                and configured_agent.agent_start_thread.is_alive()
                and self.get_exit_status()
            ):
                configured_agent.agent.preserved_state.is_aborted = True
                configured_agent.agent.advance_state(AgentState.ABORTED)

            if (
                not configured_agent.agent.current_state == AgentState.DONE
                and not configured_agent.agent.current_state == AgentState.SKIPPED
            ):
                try:
                    if configured_agent.agent_teardown_thread:
                        configured_agent.agent_teardown_thread.join(
                            timeout=self.THREAD_TIMEOUT
                        )
                    else:
                        configured_agent.agent_teardown_thread = (
                            configured_agent.agent.teardown_async()
                        )
                        self.execute_hooks(
                            "on_each_agent_teardown",
                            configured_agent.agent.config.name,
                        )
                        configured_agent.agent_teardown_thread.join(
                            timeout=self.THREAD_TIMEOUT
                        )
                    if (
                        configured_agent.agent_teardown_thread
                        and configured_agent.agent_teardown_thread.is_alive()
                    ):  # pragma: no cover
                        raise Exception(
                            f"Agent: {configured_agent.agent.config.name} Teardown step failed to complete in {self.THREAD_TIMEOUT}"
                        )
                    if self.check_for_failed_agents(configured_agent.agent):
                        raise configured_agent.agent.exception.get()
                except Exception as e:
                    self.exit_code = 1
                    configured_agent.agent.exception.put(e)
                    configured_agent.agent.advance_state(AgentState.ERROR)
                    configured_agent.agent.preserved_state.has_error = True
            self.execute_hooks(
                "on_each_agent_stop",
                configured_agent.agent.config.name,
            )

            temp_exception_queue = Queue()
            while not configured_agent.agent.exception.empty():  # print all exceptions
                error = configured_agent.agent.exception.get()
                temp_exception_queue.put(error)
                self.log.error(
                    f"Error occurred for the Agent={configured_agent.agent.config.name}",
                    exc_info=error,
                )
            configured_agent.agent.exception = temp_exception_queue

    def get_all_exceptions(self) -> list:
        """
        Get all the Exceptions occurred during the attack
        Returns:
            list of Exceptions

        """
        all_exceptions = []
        for configured_agent in self.configured_agents:
            temp_exception_queue: Queue = Queue()
            while not configured_agent.agent.exception.empty():
                e = configured_agent.agent.exception.get()
                temp_exception_queue.put(e)
                all_exceptions.append(e)
            configured_agent.agent.exception = temp_exception_queue
        return all_exceptions

    def generate_attack_report(self) -> Dict:
        """
        Generates attack report
        """

        class AgentStatus(BaseModel):
            agent_name: str
            start_time: str
            end_time: str
            status: str

        class AttackReport(BaseModel):
            """
            Attack Report Structure
            """

            id: str
            host: str
            start_time: str
            expected_end_time: str
            mode: str
            agents: List[AgentStatus]

        report: AttackReport = AttackReport(
            id=str(self.test_plan.id),
            host=os.uname()[1],
            start_time=str(self.attack_start_time),
            expected_end_time=str(self.attack_end_time),
            mode=self.test_plan.attack.mode.value,
            agents=[],
        )

        for configured_agent in self.configured_agents:
            agent = dict()
            agent["agent_name"] = configured_agent.agent.config.name
            agent["start_time"] = str(configured_agent.start_time)
            if hasattr(configured_agent.agent.config, "duration"):
                agent["end_time"] = str(configured_agent.end_time)
            else:
                agent["end_time"] = "NaN"
            if configured_agent.agent.preserved_state.has_error:
                agent["status"] = AgentState.ERROR.name
                self.exit_code = 1
            elif configured_agent.agent.preserved_state.is_aborted:
                agent["status"] = AgentState.ABORTED.name
            else:
                agent["status"] = configured_agent.agent.current_state.name
            report.agents.append(AgentStatus(**agent))
        return report.dict()

    def start_attack(self) -> int:
        """
        Performs the attack as configured in testplan
        Returns:
            attack status - 0 if successful else 1
        """
        self.log.info("Attack started")
        self.execute_hooks("on_attack_start")
        assert self.attack_end_time is not None
        assert self.configured_agents is not None
        while datetime.now(timezone.utc) <= self.attack_end_time:
            current_agent = self.get_next_agent_for_attack()
            if current_agent:
                current_agent.agent_start_thread = current_agent.agent.start_async()
                self.execute_hooks(
                    "on_each_agent_start",
                    current_agent.agent.config.name,
                )

            current_agent = self.get_next_agent_for_teardown()
            if current_agent:
                current_agent.agent_teardown_thread = (
                    current_agent.agent.teardown_async()
                )
                self.execute_hooks(
                    "on_each_agent_teardown",
                    current_agent.agent.config.name,
                )

            sleep(1)

            if self.check_for_failed_agents():
                self.exit_code = 1
                break

        self.stop_all_running_agents_in_sync()

        if self.exit_code:
            self.log.info("Attack failed")
        else:
            self.log.info("Attack Completed")
        self.execute_hooks("on_attack_completed")
        return self.exit_code



#  Copyright 2021, Yahoo
#  Licensed under the terms of the Apache 2.0 license. See the LICENSE file in the project root for terms
import time
from typing import Dict, List, Optional, Type

from pydantic import validate_arguments

from ...app_logger import AppLogger
from ...testplan import SystemState
from ...testplan.schema import TestPlan
from ...testplan.verification import VerificationConfig, VerificationType
from ...utils.hooks import EventHook
from ...utils.yaml import Dumper
from .data import VerificationData, VerificationStateData
from .plugins.BaseVerificationPlugin import BaseVerificationPlugin
from .plugins.HTTPRequestVerificationPlugin import (
    HTTPRequestVerificationPlugin,
)
from .plugins.PythonModuleVerificationPlugin import (
    PythonModuleVerificationPlugin,
)
from .plugins.SDv4VerificationPlugin import SDv4VerificationPlugin

# Enum value to corresponding Plugin Map

VERIFICATION_PLUGIN_MAP: Dict[str, Type[BaseVerificationPlugin]] = {
    "python_module": PythonModuleVerificationPlugin,
    "http_request": HTTPRequestVerificationPlugin,
    "sdv4": SDv4VerificationPlugin,
}


class VerificationController(EventHook):
    """
    Verification controller is used to run all the verification plugins configured in the testplan
    and assert that the system is expected to be in a state expected by the user. Extends the EventHook class,
    that defines the following event hooks.

    ## Valid Hooks

    === "on_start"
        Hook that gets called when the verification execution is about to start.
        No arguments are passed to the callable.

        ```python
            def callable_hook(): ...
        ```

    === "on_each_plugin_start"
        Hook that gets called when a particular plugin execution is about to start. `index` in the signature refers
        to the position in the list

        ```python
            def callable_hook(index: int, config: VerificationConfig): ...
        ```
        References:
            1. [VerificationConfig][ychaos.testplan.verification.VerificationConfig]

    === "on_each_plugin_end"
        Hook that gets called when a particular plugin execution has ended. `index` in the signature refers to the
        position in the list

        ```python
            def callable_hook(index: int, config: VerificationConfig, state_data: VerificationStateData): ...
        ```

        References:
            1. [VerificationConfig][ychaos.testplan.verification.VerificationConfig]
            2. [VerificationStateData][ychaos.core.verification.data.VerificationStateData]

    === "on_end"
        Hook that gets called when the verification execution has ended. Each element in the list
        of boolean corresponds to the result of the plugin, where `True` indicates successful verification
        and `False` is a failure to verify the state

        ```python
            def callable_hook(verify_list: List[bool]): ...
        ```

    === "on_plugin_not_found"
        Hook that gets called when a plugin available in schema is not ready for usage/not implemented.
        This case is possible for the plugins that are in Beta/development phase

        ```python
            def callable_hook(index:int, plugin_type: VerificationType): ...
        ```

    ---
    Each of the hooks get called on a certain event. The caller can register as many hooks for a particular event,
    by calling the `register_hook(event_name, hook_method)` method. All the hooks are executed sequentially. The best example
    of this is to register a hook to print information on CLI.
    """

    __hook_events__ = {
        "on_start": EventHook.CallableType(),
        "on_each_plugin_start": EventHook.CallableType(int, VerificationConfig),
        "on_each_plugin_end": EventHook.CallableType(
            int, VerificationConfig, VerificationStateData
        ),
        "on_plugin_not_found": EventHook.CallableType(int, VerificationType),
        "on_end": EventHook.CallableType(List[bool]),
    }

    @validate_arguments
    def __init__(
        self,
        testplan: TestPlan,
        current_state: SystemState,
        verification_data: List[Dict[SystemState, Optional[VerificationStateData]]],
    ):
        """
        Initialize a verification controller object.

        Args:
            testplan: A valid testplan object
            current_state: The state in which the system is expected to be in
            verification_data (List[VerificationData]): The verification data probably from previous run.
        """
        super(VerificationController, self).__init__()
        self.logger = AppLogger.get_logger(self.__class__.__name__)
        self.logger.bind(event="controller")

        self.testplan = testplan
        self.current_state = current_state

        if not verification_data:
            verification_data = [
                dict(),
            ] * len(self.testplan.verification)
        elif len(verification_data) != len(self.testplan.verification):
            raise ValueError("Data and verification config size mismatch")

        self.verification_data = list()
        for data in verification_data:
            self.verification_data.append(VerificationData.parse_obj(data))

    def execute(self) -> bool:
        """
        Execute the Verification controller.

        Returns:
            True if all the verification plugin pass, False otherwise
        """

        # Call all the hooks that were registered for `verification_start`
        # If there were no hooks registered, this will be no-op
        self.execute_hooks("on_start")

        _verify_list = list()
        for index, (verification_plugin, data) in enumerate(
            zip(self.testplan.verification, self.verification_data)
        ):

            assert isinstance(verification_plugin.states, List)  # For mypy
            if self.current_state in verification_plugin.states:
                self.logger.info(
                    msg=f"Starting {verification_plugin.type.value} verification"
                )
                plugin_class = VERIFICATION_PLUGIN_MAP.get(
                    verification_plugin.type.value, None
                )

                if plugin_class is None:
                    # This can happen when a new plugin is not implemented yet, but is
                    # available in the schema
                    self.execute_hooks(
                        "on_plugin_not_found", index, verification_plugin.type
                    )
                    continue

                plugin = plugin_class(verification_plugin.config, data)

                # Delay before verifying
                time.sleep(verification_plugin.delay_before / 1000)

                # Call all the hooks that were registered for `verification_plugin_start`.
                self.execute_hooks("on_each_plugin_start", index, verification_plugin)

                state_data = plugin.run_verification()
                self.logger.info(
                    msg=f"Completed {verification_plugin.type.value} verification"
                )

                # Call all the hooks that were registered for `verification_plugin_end`.
                self.execute_hooks(
                    "on_each_plugin_end", index, verification_plugin, state_data
                )

                data.replace_data(self.current_state, state_data)
                if verification_plugin.strict:
                    _verify_list.append(state_data.rc == 0)
            else:
                data.add_data(self.current_state, None)

            # Delay after verifying
            time.sleep(verification_plugin.delay_after / 1000)

        # Call all the hooks that were registered for `verification_end`.
        self.execute_hooks("on_end", _verify_list)

        return all(_verify_list)

    def get_encoded_verification_data(self):
        return [data.encoded_dict() for data in self.verification_data]

    def dump_verification(self, fp, output_format):
        if output_format == "json":
            self.dump_verification_json(fp)
        else:
            self.dump_verification_yaml(fp)

    def dump_verification_json(self, fp):
        import json

        json.dump(self.get_encoded_verification_data(), fp=fp, indent=4)

    def dump_verification_yaml(self, fp):
        import yaml

        yaml.dump(
            self.get_encoded_verification_data(),
            fp,
            default_flow_style=False,
            sort_keys=False,
            Dumper=Dumper,
            indent=4,
        )



#  Copyright 2021, Yahoo
#  Licensed under the terms of the Apache 2.0 license. See the LICENSE file in the project root for terms
from ...app_logger import AppLogger
from ..agent import Agent


def log_agent_lifecycle(func):
    logger = AppLogger.get_logger("agents")

    def annotation(*args, **kwargs):
        agent: Agent = args[0]
        logger.info(
            event="agents.lifecycle.start",
            agent=agent.config.name,
            method=func.__name__,
            state=agent.current_state.name,
        )
        try:
            _return_val = func(*args, **kwargs)
        except Exception as e:
            raise e from None
        finally:
            logger.info(
                event="agents.lifecycle.end",
                agent=agent.config.name,
                method=func.__name__,
                state=agent.current_state.name,
            )
        return _return_val

    return annotation



#  Copyright 2021, Yahoo
#  Licensed under the terms of the Apache 2.0 license. See the LICENSE file in the project root for terms

from argparse import Namespace
from io import StringIO

from rich.console import Console

from .main import App


class MockApp(App):
    """
    This will create a mocked application similar to [App][ychaos.cli.main.App]
    whose console output will be tracked by a StringIO object for verification. This is not
    to be used for production and only for testing purposes.
    """

    def __init__(self, args: Namespace):
        if "config" not in args:
            args.config = "dev"
        if "verbose" not in args:
            args.verbose = 0
        if "log_file" not in args:
            args.log_file = None
        if "no_color" not in args:
            args.no_color = None

        super(MockApp, self).__init__(args)
        self.console = Console(file=StringIO(), width=150)

    def get_console_output(self) -> str:
        """
        Returns the entire console output printed by the CLI application during its run

        Returns:
            String
        """
        assert isinstance(self.console.file, StringIO)
        return self.console.file.getvalue()



#  Copyright 2021, Yahoo
#  Licensed under the terms of the Apache 2.0 license. See the LICENSE file in the project root for terms
import collections
from typing import Any

from ...utils.dependency import DependencyUtils
from ...utils.hooks import EventHook

CallbackBase: Any  # For mypy

(CallbackBase,) = DependencyUtils.import_from(
    "ansible.plugins.callback", ("CallbackBase",), raise_error=False, warn=False
)
(TaskResult,) = DependencyUtils.import_from(
    "ansible.executor.task_result", ("TaskResult",), raise_error=False, warn=False
)

if CallbackBase:  # pragma: no cover

    class YChaosAnsibleResultCallback(CallbackBase, EventHook):
        """
        This is a callback plugin which help performing actions as results from running the ansible playbook comes in.
        """

        __hook_events__ = collections.defaultdict(
            EventHook.CallableType(TaskResult)
        ).fromkeys(
            (
                "on_target_unreachable",
                "on_target_failed",
                "on_target_passed",
            )
        )

        def __init__(self, *args, **kwargs):
            super(YChaosAnsibleResultCallback, self).__init__()

            EventHook.__init__(self)
            self.hooks.update(kwargs.pop("hooks", dict()))

            self.hosts_passed = dict()
            self.hosts_unreachable = dict()
            self.hosts_failed = dict()

        def v2_runner_on_unreachable(self, result):
            self.hosts_unreachable[result._host.get_name()] = result
            self.execute_hooks("on_target_unreachable", result)

        def v2_runner_on_ok(self, result):
            self.hosts_passed[result._host.get_name()] = result
            self.execute_hooks("on_target_passed", result)

        def v2_runner_on_failed(self, result, ignore_errors=False):
            self.hosts_failed[result._host.get_name()] = result
            self.execute_hooks("on_target_failed", result)



#  Copyright 2021, Yahoo
#  Licensed under the terms of the Apache 2.0 license. See the LICENSE file in the project root for terms

import shlex
import subprocess  # nosec using shlex
from enum import Enum
from ipaddress import IPv4Address, IPv4Network
from queue import LifoQueue
from typing import List, Optional, Union

from pydantic import (
    AnyHttpUrl,
    Field,
    IPvAnyAddress,
    IPvAnyNetwork,
    validate_arguments,
)

from ..agent import (
    Agent,
    AgentMonitoringDataPoint,
    AgentState,
    TimedAgentConfig,
)
from ..exceptions import AgentError
from ..utils.annotations import log_agent_lifecycle

__all__ = ["IPTablesBlockConfig", "IPTablesBlock", "DNSBlockConfig", "DNSBlock"]


class IptablesChain(Enum):
    OUTPUT = "OUTPUT"
    INPUT = "INPUT"


class IptablesRuleOperation(Enum):
    INSERT = "-I"
    DELETE = "-D"


def iptables_command_builder(
    operation, chain, port, endpoint, iptable_wait_second
) -> str:
    command = f"sudo iptables {shlex.quote(operation)} {shlex.quote(chain)}"

    args = f" -p tcp -j DROP -w {shlex.quote(str(iptable_wait_second))}"

    if port:
        args += f" --dport {shlex.quote(str(port))}"

    if endpoint:
        args += f" -d {shlex.quote(endpoint)}"

    return command + args


class IPTablesBlockConfig(TimedAgentConfig):
    name = "block_ports"
    desc = "This agent modifies the iptables rules to block traffic to specified ports or endpoint"
    is_sudo = True
    incoming_ports: Optional[List[int]] = Field(
        description="List of incoming ports to block",
        default=list(),
        examples=[[3000, 4443]],
    )
    iptables_wait: int = Field(
        description=(
            "Wait for the lock in seconds. To prevent multiple instances of the program from running concurrently, "
            "an attempt will be made to obtain an exclusive lock at launch"
        ),
        default=3,
        examples=[500, 450, 120],
        lt=1800,
        gt=0,
    )

    destination_ports: Optional[List[int]] = Field(
        description="List of destination ports to block",
        default=list(),
        examples=[[3000, 4443]],
    )

    incoming_endpoints: List[Union[IPvAnyNetwork, AnyHttpUrl, IPvAnyAddress]] = Field(
        description="List of incoming endpoint to block",
        default=list(),
        examples=["203.0.113.0", "https://yahoo.com:443"],
    )

    outgoing_endpoints: Optional[
        List[Union[IPvAnyNetwork, AnyHttpUrl, IPvAnyAddress]]
    ] = Field(
        description="List of outgoing endpoint to block",
        default=list(),
        examples=["203.0.113.0", "https://yahoo.com:443"],
    )


class IPTablesBlock(Agent):
    @validate_arguments
    def __init__(self, config: IPTablesBlockConfig):
        super(IPTablesBlock, self).__init__(config)

    def monitor(self) -> LifoQueue:
        super(IPTablesBlock, self).monitor()
        self._status.put(
            AgentMonitoringDataPoint(
                data=dict(),
                state=self.current_state,
            )
        )
        return self._status

    @log_agent_lifecycle
    def setup(self) -> None:
        super(IPTablesBlock, self).setup()

    @staticmethod
    def raise_io_error_on_iptables_failure(proc: subprocess.CompletedProcess, message):
        if proc.returncode != 0:
            raise IOError(message)

    @log_agent_lifecycle
    def run(self) -> None:
        super(IPTablesBlock, self).run()
        for port in self.config.incoming_ports:
            proc = subprocess.run(  # nosec using shlex
                iptables_command_builder(
                    IptablesRuleOperation.INSERT.value,
                    IptablesChain.INPUT.value,
                    port,
                    None,
                    self.config.iptables_wait,
                ).split(),
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
            self.raise_io_error_on_iptables_failure(
                proc, f"Error While Adding IpTable Rule: DROP {port} to INPUT Chain"
            )
        for port in self.config.destination_ports:
            proc = subprocess.run(  # nosec using shlex
                iptables_command_builder(
                    IptablesRuleOperation.INSERT.value,
                    IptablesChain.OUTPUT.value,
                    port,
                    None,
                    self.config.iptables_wait,
                ).split(),
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
            self.raise_io_error_on_iptables_failure(
                proc, f"Error While Adding IpTable Rule: DROP {port} to OUTPUT Chain"
            )

        for endpoint in self.config.incoming_endpoints:
            if isinstance(endpoint, AnyHttpUrl):
                proc = subprocess.run(  # nosec using shlex
                    iptables_command_builder(
                        IptablesRuleOperation.INSERT.value,
                        IptablesChain.INPUT.value,
                        endpoint.port,
                        str(endpoint.host),
                        self.config.iptables_wait,
                    ).split(),
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                )
                self.raise_io_error_on_iptables_failure(
                    proc,
                    f"Error While Adding IpTable Rule: DROP {endpoint} to INPUT Chain",
                )

            elif isinstance(endpoint, (IPv4Network, IPv4Address)):
                proc = subprocess.run(  # nosec using shlex
                    iptables_command_builder(
                        IptablesRuleOperation.INSERT.value,
                        IptablesChain.INPUT.value,
                        None,
                        str(endpoint),
                        self.config.iptables_wait,
                    ).split(),
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                )
                self.raise_io_error_on_iptables_failure(
                    proc,
                    f"Error While Adding IpTable Rule: DROP {endpoint} to INPUT Chain",
                )

        for endpoint in self.config.outgoing_endpoints:
            if isinstance(endpoint, AnyHttpUrl):
                proc = subprocess.run(  # nosec using shlex
                    iptables_command_builder(
                        IptablesRuleOperation.INSERT.value,
                        IptablesChain.OUTPUT.value,
                        endpoint.port,
                        str(endpoint.host),
                        self.config.iptables_wait,
                    ).split(),
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                )
                self.raise_io_error_on_iptables_failure(
                    proc,
                    f"Error While Adding IpTable Rule: DROP {endpoint} to INPUT Chain",
                )

            elif isinstance(endpoint, (IPv4Network, IPv4Address)):
                proc = subprocess.run(  # nosec using shlex
                    iptables_command_builder(
                        IptablesRuleOperation.INSERT.value,
                        IptablesChain.OUTPUT.value,
                        None,
                        str(endpoint),
                        self.config.iptables_wait,
                    ).split(),
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                )

                self.raise_io_error_on_iptables_failure(
                    proc,
                    f"Error While Adding IpTable Rule: DROP {endpoint} to INPUT Chain",
                )

    @log_agent_lifecycle
    def teardown(self) -> None:
        super(IPTablesBlock, self).teardown()
        error = False
        for port in self.config.incoming_ports:
            proc = subprocess.run(  # nosec using shlex
                iptables_command_builder(
                    IptablesRuleOperation.DELETE.value,
                    IptablesChain.INPUT.value,
                    port,
                    None,
                    self.config.iptables_wait,
                ).split(),
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
            error = proc.returncode != 0 or error

        for port in self.config.destination_ports:
            proc = subprocess.run(  # nosec using shlex
                iptables_command_builder(
                    IptablesRuleOperation.DELETE.value,
                    IptablesChain.OUTPUT.value,
                    port,
                    None,
                    self.config.iptables_wait,
                ).split(),
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
            error = proc.returncode != 0 or error

        for endpoint in self.config.incoming_endpoints:
            if isinstance(endpoint, AnyHttpUrl):
                proc = subprocess.run(  # nosec using shlex
                    iptables_command_builder(
                        IptablesRuleOperation.DELETE.value,
                        IptablesChain.INPUT.value,
                        endpoint.port,
                        str(endpoint.host),
                        self.config.iptables_wait,
                    ).split(),
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                )
                error = proc.returncode != 0 or error

            elif isinstance(endpoint, (IPv4Network, IPv4Address)):
                proc = subprocess.run(  # nosec using shlex
                    iptables_command_builder(
                        IptablesRuleOperation.DELETE.value,
                        IptablesChain.OUTPUT.value,
                        None,
                        str(endpoint),
                        self.config.iptables_wait,
                    ).split(),
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                )
                error = proc.returncode != 0 or error

        if error:
            raise AgentError("Error Occurred while removing IpTable rule")


class DNSBlockConfig(TimedAgentConfig):
    name = "block_dns"
    desc = "This agent modifies the iptables rules to block traffic to DNS ports"

    is_sudo = True

    iptables_wait: int = Field(
        description=(
            "The duration(in secs) for the agent waits to achieve the wait for the iptables command to achieve exclusive lock. "
            "This corresponds to -w option in iptables command"
        ),
        default=3,
        lt=60,
        gt=0,
    )


class DNSBlock(Agent):

    DNS_PORT = 53

    @validate_arguments
    def __init__(self, config: DNSBlockConfig):
        super(DNSBlock, self).__init__(config)

    def monitor(self) -> LifoQueue:
        return self._status

    @log_agent_lifecycle
    def setup(self) -> None:
        super(DNSBlock, self).setup()

    @staticmethod
    def raise_io_error_on_iptables_failure(proc: subprocess.CompletedProcess, message):
        if proc.returncode != 0:
            raise IOError(message)

    @log_agent_lifecycle
    def run(self):
        super(DNSBlock, self).run()

        _cmd = f"sudo iptables -I OUTPUT -p udp --dport {self.DNS_PORT} -j DROP -w {shlex.quote(str(self.config.iptables_wait))}".split()
        proc = subprocess.run(  # nosec
            _cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        self.raise_io_error_on_iptables_failure(
            proc, "Error While Adding IPTables Rule: DROP udp port: 53 to OUTPUT chain"
        )

        _cmd = f"sudo iptables -I OUTPUT -p tcp --dport {self.DNS_PORT} -j DROP -w {shlex.quote(str(self.config.iptables_wait))}".split()
        proc = subprocess.run(  # nosec
            _cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        self.raise_io_error_on_iptables_failure(
            proc, "Error While Adding IpTable Rule: DROP tcp port: 53 to OUTPUT Chain"
        )

    @log_agent_lifecycle
    def teardown(self) -> None:
        _current_state_temporary = self.current_state
        super(DNSBlock, self).teardown()
        error = False
        if _current_state_temporary in (
            AgentState.RUNNING,
            AgentState.ERROR,
            AgentState.ABORTED,
        ):
            _cmd = f"sudo iptables -D OUTPUT -p udp --dport {self.DNS_PORT} -j DROP -w {self.config.iptables_wait}".split()
            proc = subprocess.run(  # nosec
                _cmd,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )

            error = proc.returncode != 0 or error

            _cmd = f"sudo iptables -D OUTPUT -p tcp --dport {self.DNS_PORT} -j DROP -w {self.config.iptables_wait}".split()
            proc = subprocess.run(  # nosec
                _cmd,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )

            error = proc.returncode != 0 or error

        if error:
            raise AgentError("Error Occurred while removing iptables rule")



#  Copyright 2021, Yahoo
#  Licensed under the terms of the Apache 2.0 license. See the LICENSE file in the project root for terms

__copyright__: str = "Copyright 2021, Yahoo"

__import__("pkg_resources").declare_namespace(__name__)



#  Copyright 2021, Yahoo
#  Licensed under the terms of the Apache 2.0 license. See the LICENSE file in the project root for terms
from . import YChaosError


class YChaosTargetConfigConditionFailedError(YChaosError):
    def __init__(self, message: str, **kwargs):
        super(YChaosTargetConfigConditionFailedError, self).__init__(
            ValueError("Testplan target configuration condition failed"),
            message,
            **kwargs,
        )



#  Copyright 2021, Yahoo
#  Licensed under the terms of the Apache 2.0 license. See the LICENSE file in the project root for terms

from datetime import datetime, timedelta, timezone
from pathlib import Path
from queue import LifoQueue
from socket import gaierror, timeout
from types import SimpleNamespace
from typing import List, Union

from pydantic import (
    AnyHttpUrl,
    BaseModel,
    Field,
    FilePath,
    validate_arguments,
    validator,
)

from ...utils.builtins import AEnum
from ...utils.dependency import DependencyUtils
from ..agent import Agent, AgentConfig, AgentMonitoringDataPoint, AgentPriority
from ..utils.annotations import log_agent_lifecycle

(X509,) = DependencyUtils.import_from(
    "OpenSSL.crypto", ("X509",), raise_error=False, warn=False
)

pyopenssl = DependencyUtils.import_module(
    "requests.packages.urllib3.contrib.pyopenssl", raise_error=False, warn=False
)


class ServerCertValidationConfig(AgentConfig):
    name = "server_cert_validation"
    desc = "This agent retrieves SSL certificates from server and validates it"
    priority = AgentPriority.LOW_PRIORITY

    urls: List[AnyHttpUrl] = Field(
        default=list(),
        description="List of URLs which will be validated for their SSL certificate expiry",
        min_items=1,
    )
    expiry_threshold: timedelta = Field(
        default=timedelta(days=7), description="Expiry threshold"
    )

    timeout: int = Field(
        default=5, description="Default timeout to fetch the certificates in seconds"
    )


class ServerCertValidation(Agent):
    @validate_arguments
    def __init__(self, config: ServerCertValidationConfig):
        super(ServerCertValidation, self).__init__(config)

    def monitor(self) -> LifoQueue:
        super(ServerCertValidation, self).monitor()
        return self._status

    @log_agent_lifecycle
    def setup(self) -> None:
        super(ServerCertValidation, self).setup()

    @staticmethod
    def get_server_cert(host: str, port: int, timeout_=5):
        from socket import setdefaulttimeout

        setdefaulttimeout(timeout_)

        assert pyopenssl is not None
        return pyopenssl.OpenSSL.crypto.load_certificate(
            pyopenssl.OpenSSL.crypto.FILETYPE_PEM,
            pyopenssl.ssl.get_server_certificate((host, port)),
        )

    @log_agent_lifecycle
    def run(self) -> None:
        for url in self.config.urls:
            if url.port is None:
                url.port = "443"

            data = dict(host=url.host, port=int(url.port), error=None)

            try:
                cert = self.get_server_cert(
                    url.host, int(url.port), timeout_=self.config.timeout
                )
                cert_expiry_date = datetime.strptime(
                    cert.get_notAfter().decode("ascii"), "%Y%m%d%H%M%SZ"
                )
                cert_expiry_date.replace(tzinfo=timezone.utc)

                data.update(
                    dict(
                        not_valid_after=cert_expiry_date,
                        is_expired=datetime.utcnow() >= cert_expiry_date,
                        is_critical=datetime.utcnow() + self.config.expiry_threshold
                        >= cert_expiry_date,
                    )
                )
            except (timeout, gaierror) as error:
                data.update(error=str(error.__class__.__name__))

            self._status.put(
                AgentMonitoringDataPoint(
                    data=data,
                    state=self.current_state,
                )
            )

    @log_agent_lifecycle
    def teardown(self) -> None:
        super(ServerCertValidation, self).teardown()


class CertificateFileType(AEnum):
    """
    Attributes:
        ASN1: (asn1) ASN1 Certificate format
        PEM: (pem) PEM Certificate format
    """

    ASN1 = "asn1", SimpleNamespace()
    PEM = "pem", SimpleNamespace()

    def binder(self):
        pyopenssl = DependencyUtils.import_module(
            "requests.packages.urllib3.contrib.pyopenssl", raise_error=False
        )
        if self == self.PEM:
            return pyopenssl.OpenSSL.crypto.FILETYPE_PEM
        elif self == self.ASN1:
            return pyopenssl.OpenSSL.crypto.FILETYPE_ASN1
        else:  # pragma: no cover
            raise Exception("Unknown Certificate FileType")


class CertificateFileConfig(BaseModel):
    path: FilePath = Field(..., description="Path to the certificate file")
    type: CertificateFileType = Field(
        default=CertificateFileType.PEM, description="Type of the certificate file"
    )


class CertificateFileValidationConfig(AgentConfig):
    name = "cert_file_validation"
    desc = "This agent decodes the local certificates and validates for expiry/critical"
    priority = AgentPriority.LOW_PRIORITY

    expiry_threshold: timedelta = Field(
        default=timedelta(days=7), description="Expiry threshold"
    )

    paths: List[Union[FilePath, CertificateFileConfig]] = Field(
        default=list(),
        description="List of Certificate files to be validated",
        min_items=1,
    )

    @validator("paths", each_item=True)
    def parse_paths(cls, v, values):
        if isinstance(v, Path):
            return CertificateFileConfig(path=v)
        return v


class CertificateFileValidation(Agent):
    def monitor(self) -> LifoQueue:
        super(CertificateFileValidation, self).monitor()
        return self._status

    @log_agent_lifecycle
    def setup(self) -> None:
        super(CertificateFileValidation, self).setup()

    @log_agent_lifecycle
    def run(self) -> None:
        super(CertificateFileValidation, self).run()
        for cert_path_config in self.config.paths:
            data = dict(
                path=str(cert_path_config.path),
                type=cert_path_config.type.value,
                error=None,
            )
            try:
                assert pyopenssl is not None
                cert = pyopenssl.OpenSSL.crypto.load_certificate(
                    cert_path_config.type.binder(),
                    cert_path_config.path.read_bytes(),
                )

                cert_expiry_date = datetime.strptime(
                    cert.get_notAfter().decode("ascii"), "%Y%m%d%H%M%SZ"
                )
                cert_expiry_date.replace(tzinfo=timezone.utc)

                data.update(
                    dict(
                        not_valid_after=cert_expiry_date,
                        is_expired=datetime.utcnow() >= cert_expiry_date,
                        is_critical=datetime.utcnow() + self.config.expiry_threshold
                        >= cert_expiry_date,
                    )
                )
            except pyopenssl.OpenSSL.crypto.Error:  # type: ignore
                data.update(error="decoding_error")

            self._status.put(
                AgentMonitoringDataPoint(
                    data=data,
                    state=self.current_state,
                )
            )

    @log_agent_lifecycle
    def teardown(self) -> None:
        super(CertificateFileValidation, self).teardown()



#  Copyright 2021, Yahoo
#  Licensed under the terms of the Apache 2.0 license. See the LICENSE file in the project root for terms

from types import SimpleNamespace

from ..utils.builtins import AEnum
from .contrib import ContribAgentConfig
from .network.iptables import (
    DNSBlock,
    DNSBlockConfig,
    IPTablesBlock,
    IPTablesBlockConfig,
)
from .network.traffic import TrafficBlock, TrafficBlockConfig
from .special.NoOpAgent import (
    NoOpAgent,
    NoOpAgentConfig,
    NoOpTimedAgent,
    NoOpTimedAgentConfig,
)
from .system.cpu import CPUBurn, CPUBurnConfig
from .system.disk import DiskFill, DiskFillConfig
from .system.icmp import PingDisable, PingDisableConfig
from .validation.certificate import (
    CertificateFileValidation,
    CertificateFileValidationConfig,
    ServerCertValidation,
    ServerCertValidationConfig,
)

__all__ = ["AgentType"]


class AgentType(AEnum):

    # Special Agents
    NO_OP = "no_op", SimpleNamespace(schema=NoOpAgentConfig, agent_defn=NoOpAgent)
    NO_OP_TIMED = "no_op_timed", SimpleNamespace(
        schema=NoOpTimedAgentConfig, agent_defn=NoOpTimedAgent
    )

    # System Agents
    CPU_BURN = "cpu_burn", SimpleNamespace(schema=CPUBurnConfig, agent_defn=CPUBurn)

    # Network Agents
    IPTABLES_BLOCK = "iptables_block", SimpleNamespace(
        schema=IPTablesBlockConfig, agent_defn=IPTablesBlock
    )
    DNS_BLOCK = "dns_block", SimpleNamespace(schema=DNSBlockConfig, agent_defn=DNSBlock)
    TRAFFIC_BLOCK = "traffic_block", SimpleNamespace(
        schema=TrafficBlockConfig, agent_defn=TrafficBlock
    )

    # Validation Agents
    SERVER_CERT_VALIDATION = "server_cert_validation", SimpleNamespace(
        schema=ServerCertValidationConfig, agent_defn=ServerCertValidation
    )
    CERT_FILE_VALIDATION = "cert_file_validation", SimpleNamespace(
        schema=CertificateFileValidationConfig, agent_defn=CertificateFileValidation
    )

    # Special Contrib agent
    CONTRIB = "contrib", SimpleNamespace(  # pragma: no cover
        schema=ContribAgentConfig, agent_defn=lambda config: config.get_agent()
    )

    DISABLE_PING = "disable_ping", SimpleNamespace(
        schema=PingDisableConfig, agent_defn=PingDisable
    )

    DISK_FILL = "disk_fill", SimpleNamespace(schema=DiskFillConfig, agent_defn=DiskFill)
